/* ------------------------------------------------------------------ */
/*		     S o r t V e c t o r . j a v a		      */
/*		     =============================		      */
/*								      */
/*   This file contains the source of the "SortVector" abstract       */
/*   class.							      */
/* ------------------------------------------------------------------ */


import java.util.*;


/**
 * This abstract class is used to sort values into a vector. Classes
 * extending this class must implement the <CODE>compare&nbsp;()</CODE>
 * method.
 *
 *    @author  Hans Blom
 *
******************************************/
abstract class  SortVector  extends  Vector
/*****************************************/
{
    /*
     * Set the current index in the vector which was last used during
     * the last call.
     */

    private int  curr_ind = 0;


    /**
     * This function must be implemented by classes extending this class
     * and is used to sort a new value into the vector. The result of
     * the comparison must be returned as:
     *
     * <UL>
     *
     * <LI><B>&lt;&nbsp;0</B> when <CODE>object_1</CODE> should be
     *     placed after <CODE>object_1</CODE>.</LI>
     *
     * <LI><B>&gt;&nbsp;0</B> when <CODE>object_1</CODE> should be
     *     placed before <CODE>object_1</CODE>.</LI>
     *
     * <LI><B>zero</B> when both objects are zero.</LI>
     *
     * </UL>
     *
     *    @param  object_1   The first object to compare.
     *
     *    @param  object_2   The second object to compare.
     *
     *    @return  The result of the comparison.
     *
    ****************************************/
    abstract int  compare (Object  object_1,
			   Object  object_2);
    /***************************************/


    /**
     * Clear the vector.
     *
    *************/
    void  clear ()
    /************/
    {
	removeAllElements ();    // Remove all elements from the vector.
	curr_ind = 0;		 // Reset the index of the current elem.
    }


    /**
     * Sort a new element into the vector. It is only added to the
     * vector when it was not already there. A value of
     * <CODE>true</CODE> is returned when the element is added. Else a
     * value of <CODE>false</CODE> is returned.
     *
     *    @param  new_elem   The value to sort into the array.
     *
     *    @return  The result if the element could be added.
     *
    ********************************************/
    boolean  sortInsertElement (Object  new_elem)
    /*******************************************/
    {
	int  curr_size = size ();    // Get current size of this vector.

	if (curr_size == 0)    // Insert element as first of the vector.
	{
	    insertElementAt (new_elem, curr_ind);
	    return true;
	}

	/*
	 * Sort the element into the vector.
	 */

	boolean  inserted = false;

	/*
	 * Compare the new element with the current element of the
	 * vector.
	 */

	int  cmp_val = compare (new_elem, elementAt (curr_ind));

	if (cmp_val > 0)    // New element was "larger".
	{
	    /*
	     * Go to the right in the vector while the new element is
	     * larger.
	     */

	    while (cmp_val > 0  &&  ++curr_ind < curr_size)
		cmp_val = compare (new_elem, elementAt (curr_ind));

	    /*
	     * Insert the new element when it was not yet in the
	     * vector.
	     */

	    if (cmp_val < 0  ||  curr_ind == curr_size)
	    {
		insertElementAt (new_elem, curr_ind);
		inserted = true;
	    }
	}
	else if (cmp_val < 0)    // The new element was "smaller".
	{
	    /*
	     * Go left in the vector while the element at the left of
	     * the current element is smaller than the new element.
	     */

	    while ( curr_ind > 0  &&
		    (cmp_val = compare (new_elem,
					elementAt (curr_ind-1))) < 0 )
	    {
		curr_ind--;
	    }

	    /*
	     * Insert the new element when it was not yet in the
	     * vector.
	     */

	    if (curr_ind == 0  ||  cmp_val > 0)
	    {
		insertElementAt (new_elem, curr_ind);
		inserted = true;
	    }
	}

	return inserted;    // Return if the new element is inserted.

    } // sortInsertElement ()

} // class SortVector
