/* ------------------------------------------------------------------ */
/*		    U D P B W F o r m a t . j a v a		      */
/*		    ===============================		      */
/*								      */
/*   This file contains the source of the "UDPBWFormat" class.	      */
/* ------------------------------------------------------------------ */


/**
 * This class defined the format of an UDP bandwidth value. It contains
 * the receive data rate and percentage #&nbsp;frames lost values.
 *
 *    @author  Hans Blom
 *
****************/
class UDPBWFormat
/***************/
{
    /*
     * Define the separator between the UDP bandwidth fields.
     */

    private static final char  UDP_BW_SEPAR = '/';

    /*
     * Define the UDP bandwidth types and the # UDP bandwidth types. The
     * type ID's are also the indices of the corresponding fields.
     */

    static final int  RATE_TYPE = 0;
    static final int  LOST_TYPE = 1;

    private static final int  NR_UDP_BW_TYPES = 2;

    /*
     * Define the array to store the split UDP bandwidth fields.
     */

    private String []  udp_bw_fields;


    /**
     * Constructor. It allocates the UDP bandwidth fields array and
     * initialises the UDP bandwidth field values to zero.
     *
    *************/
    UDPBWFormat ()
    /************/
    {
	udp_bw_fields = new String [NR_UDP_BW_TYPES];

	for (int i = 0;  i < NR_UDP_BW_TYPES;  i++)
	    udp_bw_fields [i] = "0";
    }


    /**
     * Set the UDP bandwidth fields from an input string.
     *
     *    @param  udp_bw_str   The string which contains the UDP
     *			       bandwidth fields.
     *
     *    @exception  Exception   An exception is thrown when the input
     *				  UDP bandwidth string has a bad format.
     *
    ************************************************/
    void  set (String  udp_bw_str)  throws  Exception
    /***********************************************/
    {
	StringTools.split (udp_bw_fields, UDP_BW_SEPAR, udp_bw_str);
    }


    /**
     * Return the UDP bandwidth value with the specified type.
     *
     *    @param  udp_bw_type   The UDP bandwidth type.
     *
     *    @return  The UDP bandwidth type value.
     *
     *    @exception  IndexOutOfBoundsException   The specified UDP
     *						  bandwidth type is not
     *						  in the specified
     *						  range.
     *
    ****************************************************************/
    String  get (int  udp_bw_type)  throws  IndexOutOfBoundsException
    /***************************************************************/
    {
	if (udp_bw_type < 0  ||  udp_bw_type >= NR_UDP_BW_TYPES)
	{
	    throw new IndexOutOfBoundsException (
			  "UDP bandwidth type out of bounds: " +
			  udp_bw_type);
	}

	return udp_bw_fields [udp_bw_type];
    }


    /**
     * Return if an UDP bandwidth value is defined. This is the case
     * when all UDP bandwidth fields do not have
     * &quot;<CODE>-</CODE>&quot; as first character.
     *
     *    @return  The result of the test to control if the UDP
     *		   bandwidth value was defined.
     *
    ********************/
    boolean  isDefined ()
    /*******************/
    {
	boolean  defined = true;
	int	 i	 = 0;

	while (i < NR_UDP_BW_TYPES  &&  defined)
	    if (udp_bw_fields [i++].charAt (0) == '-')
		defined = false;

	return defined;
    }

} // class UDPBWFormat
