/* ------------------------------------------------------------------ */
/*		    D a t a H a n d l e r . j a v a		      */
/*		    ===============================		      */
/*								      */
/*   This file contains the source of the "DataHandler" interface.    */
/*   It should be implemented to handle the rounding of the data at   */
/*   tick lines and to get the tick labels.			      */
/* ------------------------------------------------------------------ */


package user.plot;    // Define the plot package.


/**
 * This interface should be implemented to:
 *
 * <UL>
 *
 * <LI>Handle the rounding of the data to obtain the extrema in the plot
 *     area and to get the tick lines.</LI>
 *
 * <LI>To convert the tick values to one or more tick label lines.</LI>
 *
 * </UL>
 *
 *    @author  Hans Blom
 *
***************************/
public interface DataHandler
/**************************/
{
    /**
     * This method should be implemented to round the specified value
     * within the specified range. This method will be used for axis
     * labeling.
     *
     *    @param  value   The value to round.
     *
     *    @param  range   The range in which the rounding should be
     *			  performed.
     *
     *    @return   The rounded value.
     *
     *    @see PlotCanvas#round
     *
    ************************************/
    public double  round (double  value,
			  double  range);
    /***********************************/


    /**
     * This method should be implemented to return he next tick value
     * using the specified value and step.
     * </P>
     *
     * </P>
     * Note that this method is only used with <EM>linear</EM> scaling.
     *
     *    @param  value   The current tick value.
     *
     *    @param  step    The step to the next tick value.
     *
     *    @return   The next tick value.
     *
     *    @see  PlotCanvas#nextTickValue
     *    @see  PlotCanvas#setXScaling
     *    @see  PlotCanvas#setYScaling
     *
    *******************************************/
    public double  nextTickValue (double  value,
				  double  step);
    /******************************************/

			  
    /**
     * This method should be implemented to return the #&nbsp;lines of
     * the tick labels.
     *
     *    @return   The #&nbsp;lines of the tick labels.
     *
     *    @see  PlotCanvas#getNrTickLines
     *
    *****************************/
    public int  getNrTickLines ();
    /****************************/


    /**
     * This method should be implemented to return the tick label lines
     * where the specified value is converted to the requested labels.
     * The specified step to the next tick value may be used as scaling
     * parameter.
     *
     *    @param  value   The tick value to convert.
     *
     *    @param  step    The step to the next tick value, intended for
     *                    scaling.
     *
     *    @return   The tick label lines.
     *
     *    @see  PlotCanvas#getTickLines
     *
    *********************************************/
    public String []  getTickLines (double  value,
				    double  step);
    /********************************************/

} // interface DataHandler
