/* ------------------------------------------------------------------ */
/*		     P l o t C a n v a s . j a v a		      */
/*		     =============================		      */
/*								      */
/*   This file contains the source of the "PlotCanvas" class in the   */
/*   "user.plot" package.					      */
/* ------------------------------------------------------------------ */


package user.plot;    // Define the plot package.

import java.lang.*;
import java.io.*;
import java.awt.*;
import java.util.*;
import user.common.*;


/**
 * Include high resolution graphics in JAVA.
 * </P>
 *
 * <P>
 * This package is designed for the scientific needs rather than for
 * business graphics. It will display ten thousands of data points with
 * error bars, but you will miss any fancy graphics options as stacked
 * bars in 3D look or pie charts.
 * </P>
 *
 * <P>
 * History:
 *
 * <DL>
 *
 * <DT>Version 1.1</DT>
 * <DD>The initial version from Friedrich Schotte. The package was an
 *     extension of the Appler class.</DD>
 *
 * <DT>Version 2.0</DT>
 * <DD>Revision by Hans Blom. The package is split into the following
 *     classes and interfaces:
 *     <DL>
 *     <DT><CODE>PlotCanvas</CODE></DT>
 *     <DD>This class is the plot canvas. The source is in this
 *	   file.<DD>
 *     <DT><CODE>PlotMarker</CODE></DT>
 *     <DD>This class defines the plot markers.</DD>
 *     <DT><CODE>PlotApplet</CODE></DT>
 *     <DD>This class contains the general plot Applet methods. Note
 *         that it allows the plotting of the <CODE>PlotCanvas</CODE>
 *         into a new frame.</DD>
 *     <DT><CODE>PlotDataApplet</CODE></DT>
 *     <DD>This class contains the data I/O methods.</DD>
 *     <DT><CODE>PlotTraceHandler</CODE></DT>
 *     <DD>This interface can be implemented to supply an index for a
 *	   plot trace, given a specified index. The trace index is used
 *	   to define a trace color, plot markers, etc.</DD>
 *     <DT><CODE>PlotStatusHandler</CODE></DT>
 *     <DD>This interface can be implemented to handle the status
 *	   messages.</DD>
 *     <DT><CODE>DataHandler</CODE></DT>
 *     <DD>This interface can be implemented to handle data ticks and
 *	   tick labels.</DD>
 *     </DL></DD>
 *
 *     </DL>
 *
 *    @see  PlotApplet
 *    @see  PlotMarker
 *    @see  PlotDataApplet
 *    @see  PlotTraceHandler
 *    @see  PlotStatusHandler
 *    @see  DataHandler
 *
 *    @author  Friedrich Schotte
 *    @author  Hans Blom
 *
 *    @version  2.0, January 1999 (for Java 1.0)
 *
*******************************************************/
public class  PlotCanvas  extends     Canvas
			  implements  Runnable,
			  	      PlotTraceHandler,
				      PlotStatusHandler,
				      DataHandler
/******************************************************/
{
    /*
     * Define the public constants used for coordinate scaling.
     */

    /**
     * Use a linear scaling for the <I>X</I>, or
     * <I>Y</I>&nbsp;coordinate.
     */

    public static final int  LINEAR = 0;

    /**
     * Use a logarithmic scaling for the <I>X</I>, or
     * <I>Y</I>&nbsp;coordinate.
     */

    public static final int  LOG = 1;


    /*
     * Define the public constants used for status messages at (action)
     * events.
     */

    /**
     * Enter the plot canvas.
     */

    public static final int  ENTER = 0;

    /**
     * Stay within the plot canvas.
     */

    public static final int  WITHIN = 1;

    /**
     * Exit the plot canvas.
     */

    public static final int  EXIT = 2;


    /*
     * Define the background and foreground colors of the plot canvas,
     * used during initialisation.
     */

    private static final Color  BACKGROUND = Color.white;
    private static final Color  FOREGROUND = Color.black;

    /*
     * Define the outer border around the plot area.
     */

    private static final int  OUTER_AREA_BORDER = 5;

    /*
     * Define the minimal zoom length.
     */

    private static final int  ZOOM_LENGTH_BOUND = 5;

    /*
     * Define the default # lines in a tick label.
     */

    private static final int  NR_DATA_LINES = 1;

    /*
     * Define the minimum relative value for an accepted step.
     */

    private static final double  REL_STEP_BOUND = 1.0E-10;

    /*
     * Define the maximum relative step a rounded value may be shifted
     * according to the rounding range.
     */

    private static final double  REL_MAX_RANGE_STEP = 0.9;

    /*
     * Define the value used for writing scaled doubles.
     */

    private static final double  VALUE_SCALE = 1000.0;

    /*
     * Define log(10) as a constant.
     */

    private static final double  LOG_10 = Math.log (10.0);


    /*
     * Declare the class instances.
     */

    private String []  data_lines;     // Storage tick label lines.
    private Dimension  canvas_size;    // Contains current canvas size.
    private Rectangle  area;	       // The area to plot the points.
    private double     xmin;	       // X minimum all data.
    private double     xmax;	       // X maximum ,,   ,, .
    private double     ymin;	       // Y minimum ,,   ,, .
    private double     ymax;	       // Y maximum ,,   ,, .
    private double     xmin_zoom;      // X minimum data zoom region.
    private double     xmax_zoom;      // X maximum  ,,   ,,    ,,  .
    private double     ymin_zoom;      // Y minimum  ,,   ,,    ,,  .
    private double     ymax_zoom;      // Y maximum  ,,   ,,    ,,  .
    private boolean    zoomed;	       // Zoomed into the plot?
    private Double     xmin_obj;       // Value from applet parameters.
    private Double     xmax_obj;       //   ,,   ,,    ,,       ,,
    private Double     ymin_obj;       //   ,,   ,,    ,,       ,,
    private Double     ymax_obj;       //   ,,   ,,    ,,       ,,
    private Rectangle  selection;      // The selected zoom rectangle.
    private Point      mouse_down;     // Position moue went down.
    private boolean    mouse_drag;     // Dragging mouse?
    private int	       x_tick_spacing; // Spacing X ticks.
    private int	       y_tick_spacing; // Spacing Y ticks.
    private int        tick_length;    // Length tick line.

    /*
     * Data is grouped into traces of data points. The entries in the
     * "data" vector refer to the "traces", which are vectors containing
     * elements of type  "DataPoint".
     */

    private Vector  data;

    /*
     * Define the thread variables used in the paint thread: this
     * thread does all the painting. This is done to make it possible
     * to interrupt the painting by pending events. This is especially
     * useful for nervous users at slow displays.
     */

    private Graphics  paint_graphics;  // Graphics for the paint thread.
    private Thread    paint_thread;    // Paint thread to paint canvas.
    private QuitSync  paint_quit;      // Quit from paint thread?

    /*
     * Define the handler to return a plot trace index, given an input
     * index.
     */

    private PlotTraceHandler  trace_handler;

    /*
     * Define the handler of the plot status messages. Default the
     * messages are ignored by the "showMessage()" method in this class.
     */

    private PlotStatusHandler  status_handler;

    /*
     * The following variables can be set to an interface, implemented
     * by the user, to handle the x and/or y data. Default the
     * implementation in this class are used.
     */

    private DataHandler  x_handler;
    private DataHandler  y_handler;

    /*
     * More general instance variables to come.
     */

    private Vector   x_ticks;	      // Vector with X-tick coordinates.
    private Vector   y_ticks;	      // Vector with Y-tick coordinates.
    private Vector   colors;	      // Vector with plot trace colors.
    private Color    background;      // Background color of plot area.
    private String   x_axis_label;    // Label X-axis.
    private String   y_axis_label;    // Label Y-axis.
    private Vector   x_offsets;	      // Vector X-offsets plot traces.
    private Vector   y_offsets;	      // Vector Y-offsets plot traces.
    private Vector   join;	      // Vector join values plot traces.
    private boolean  use_join;	      // For all traces do join points?
    private Vector   markers;	      // Vector used markers in traces.
    private boolean  use_markers;     // For all traces do mark points?
    private int      x_scaling;	      // Scaling type X-coordinates.
    private int      y_scaling;       // Scaling type Y-coordinates.


    /**
     * Constructor.
     *
    ********************/
    public  PlotCanvas ()
    /*******************/
    {
	/*
	 * Initialise the class instances.
	 */

	data_lines     = new String [NR_DATA_LINES];
	canvas_size    = new Dimension (0, 0);
	area           = new Rectangle ();
	xmin           = 0.1;
	xmax           = 1.0;
	ymin           = 0.1;
	ymax           = 1.0;
	zoomed	       = false;
	xmin_obj       = null;
	xmax_obj       = null;
	ymin_obj       = null;
	ymax_obj       = null;
	selection      = null;
	mouse_down     = null;
	mouse_drag     = false;
	x_tick_spacing   = 60;
	y_tick_spacing   = 60;
	tick_length    = 4;
	data           = new Vector ();
	paint_thread   = null;
	paint_quit     = new QuitSync ();
	trace_handler  = this;
	status_handler = this;
	x_handler      = this;
	y_handler      = this;
	x_ticks        = new Vector ();
	y_ticks        = new Vector ();
	colors         = new Vector ();
	background     = null;
	x_axis_label   = null;
	y_axis_label   = null;
	x_offsets      = new Vector ();
	y_offsets      = new Vector ();
	join           = new Vector ();
	use_join       = true;
	markers        = new Vector ();
	use_markers    = true;
	x_scaling      = LINEAR;
	y_scaling      = LINEAR;

	/*
	 * Initialise the background and foreground colors of the plot
	 * canvas.
	 */

	setBackground (BACKGROUND);
	setForeground (FOREGROUND);

    } // PlotCanvas ()


    /**
     * Set the label of the <I>X</I>-axis.
     *
     *    @param  x_axis_label   The specified label of the
     *				 <I>X</I>-axis.
     *
    ************************************************/
    public void  setXAxisLabel (String  x_axis_label)
    /***********************************************/
    {
	this.x_axis_label = x_axis_label;
    }


    /**
     * Get the label of the <I>X</I>-axis.
     *
     *    @return   The label of the <I>X</I>-axis.
     *
    ******************************/
    public String  getXAxisLabel ()
    /*****************************/
    {
	return x_axis_label;
    }


    /**
     * Set the label of the <I>Y</I>-axis.
     *
     *    @param  y_axis_label   The specified label of the
     *				 <I>Y</I>-axis.
     *
    ************************************************/
    public void  setYAxisLabel (String  y_axis_label)
    /***********************************************/
    {
	this.y_axis_label = y_axis_label;
    }


    /**
     * Get the label of the <I>Y</I>-axis.
     *
     *    @return   The label of the <I>Y</I>-axis.
     *
    ******************************/
    public String  getYAxisLabel ()
    /*****************************/
    {
	return y_axis_label;
    }


    /**
     * Set the interface to supply an index for a plot trace, given the
     * specified index. Default the specified index is used.
     *
     *    @param  trace_handler   The specified handler to supply a
     *				  trace index.
     *
     *    @see  PlotTraceHandler
     *
    *****************************************************************/
    public void  setPlotTraceHandler (PlotTraceHandler  trace_handler)
    /****************************************************************/
    {
	this.trace_handler = trace_handler;
    }


    /**
     * Get the interface which supplies an index for a plot trace, given
     * the specified index.
     *
     *    @return   The interface which supplies a trace index.
     *
     *    @see  PlotTraceHandler
     *
    **********************************************/
    public PlotTraceHandler  getPlotTraceHandler ()
    /*********************************************/
    {
	return trace_handler;
    }


    /**
     * Set the interface which handles the plot status messages. Default
     * these messages are ignored by the implementation of this class.
     *
     *    @param  status_handler   The specified handler of the status
     *				   messages.
     *
     *    @see  PlotStatusHandler
     *
    *******************************************************************/
    public void  setPlotStatusHandler (PlotStatusHandler status_handler)
    /******************************************************************/
    {
	this.status_handler = status_handler;
    }


    /**
     * Get the interface which handles the status messages.
     *
     *    @return   The interface which handles the status messages.
     *
     *    @see  PlotStatusHandler
     *
    ************************************************/
    public PlotStatusHandler  getPlotStatusHandler ()
    /***********************************************/
    {
	return status_handler;
    }


    /**
     * Set the interface which handles the rounding of the
     * <I>X</I>-data. The default implementation by this class is
     * sufficient for &quot;normal&quot; double values.
     *
     *    @param  x_handler   The specified rounding interface.
     *
     *    @see DataHandler
     *
    ****************************************************/
    public void  setXDataHandler (DataHandler  x_handler)
    /***************************************************/
    {
	this.x_handler = x_handler;
    }


    /**
     * Get the interface which handles the rounding of the
     * <I>X</I>-data.
     *
     *    @return   The interface which handles the rounding of the
     *		    <I>X</I>-data.
     *
     *    @see  DataHandler
     *
    *************************************/
    public DataHandler  getXDataHandler ()
    /************************************/
    {
	return x_handler;
    }


    /**
     * Set the interface which handles the rounding of the
     * <I>Y</I>-data. The default implementation by this class is
     * sufficient for &quot;normal&quot; double values.
     *
     *    @param  y_handler   The specified rounding interface.
     *
     *    @see DataHandler
     *
    ****************************************************/
    public void  setYDataHandler (DataHandler  y_handler)
    /***************************************************/
    {
	this.y_handler = y_handler;
    }


    /**
     * Get the interface which handles the rounding of the
     * <I>Y</I>-data.
     *
     *    @return   The interface which handles the rounding of the
     *		    <I>Y</I>-data.
     *
     *    @see  DataHandler
     *
    *************************************/
    public DataHandler  getYDataHandler ()
    /************************************/
    {
	return y_handler;
    }


    /**
     * Set the length of the tick marks.
     *
     *    @param  tick_length   The specified length of the tick marks.
     *
    ********************************************/
    public void  setTickLength (int  tick_length)
    /*******************************************/
    {
	this.tick_length = tick_length;
    }


    /**
     * Get the length of the tick marks.
     *
     *    @return   The length of the tick marks.
     *
    ***************************/
    public int  getTickLength ()
    /**************************/
    {
	return tick_length;
    }


    /**
     * Set the spacing between the ticks at the <I>X</I>-axis.
     *
     *    @param  x_tick_spacing   The specified tick spacing.
     *
    *************************************************/
    public void  setXTickSpacing (int  x_tick_spacing)
    /************************************************/
    {
	this.x_tick_spacing = x_tick_spacing;
    }


    /**
     * Get the spacing between the ticks at the <I>X</I>-axis.
     *
     *    @return   The tick spacing.
     *
    *****************************/
    public int  getXTickSpacing ()
    /****************************/
    {
	return x_tick_spacing;
    }


    /**
     * Set the spacing between the ticks at the <I>Y</I>-axis.
     *
     *    @param  y_tick_spacing   The specified tick spacing.
     *
    *************************************************/
    public void  setYTickSpacing (int  y_tick_spacing)
    /************************************************/
    {
	this.y_tick_spacing = y_tick_spacing;
    }


    /**
     * Get the spacing between the ticks at the <I>Y</I>-axis.
     *
     *    @return   The tick spacing.
     *
    *****************************/
    public int  getYTickSpacing ()
    /****************************/
    {
	return y_tick_spacing;
    }


    /**
     * Add a color to the vector with trace colors.
     *
     *    @param  color   The specified trace color.
     *
    ****************************************/
    public void  addTraceColor (Color  color)
    /***************************************/
    {
	colors.addElement (color);
    }


    /**
     * Remove all colors from the vector with trace colors.
     *
     *    @see  PlotCanvas#removeTraceColor
     *    @see  PlotCanvas#removeTraceColorAt
     *
    ***********************************/
    public void  removeAllTraceColors ()
    /**********************************/
    {
	colors.removeAllElements ();
    }


    /**
     * Remove the specified color from the vector with trace colors.
     *
     *    @param  color   The trace color to remove
     *
     *    @return   <CODE>true</CODE>if the argument was a component of
     *		    the vector with trace colors; <CODE>false</CODE>
     *		    otherwise
     *
     *    @see  PlotCanvas#removeAllTraceColors
     *    @see  PlotCanvas#removeTraceColorAt
     *
    **********************************************/
    public boolean  removeTraceColor (Color  color)
    /*********************************************/
    {
	return colors.removeElement (color);
    }


    /**
     * Remove the specified color from the vector with trace colors at
     * the specified index.
     *
     *    @param  index   The index of the trace color to remove
     *
     *    @see  PlotCanvas#removeAllTraceColors
     *    @see  PlotCanvas#removeTraceColor
     *
    *******************************************/
    public void  removeTraceColorAt (int  index)
    /******************************************/
    {
	colors.removeElementAt (index);
    }


    /**
     * Get the #&nbsp;colors stored in the vector with trace colors.
     *
     *    @return   The #&nbsp;trace colors.
     *
    ******************************/
    public int  getNrTraceColors ()
    /*****************************/
    {
	return colors.size ();
    }


    /**
     * Get the trace color with the specified index from the trace color
     * vector. The foreground color of the <CODE>PlotCanvas</CODE> is
     * returned when the vector is empty. If the index is larger than
     * the available #&nbsp;colors stored (<CODE>nr_colors</CODE>), the
     * color with index <CODE>index&nbsp;% nr_colors</CODE> is returned.
     *
     *    @param  index   The specified index of the trace color.
     *
     *    @return   The trace color.
     *
    ***************************************/
    public Color  getTraceColor (int  index)
    /**************************************/
    {
	int  nr_colors = colors.size ();

	if (nr_colors == 0)
	    return getForeground ();
	else
	    return (Color) colors.elementAt (index % nr_colors);
    }


    /**
     * Set the background color of the area where the data are plotted.
     * The default is the background of the <CODE>PlotCanvas</CODE>.
     *
     *    @param  background   The specified plot area background.
     *
    *************************************************/
    public void  setAreaBackground (Color  background)
    /************************************************/
    {
	this.background = background;
    }


    /**
     * Get the background color of the area where the data are plotted.
     * The default is the background of the <CODE>PlotCanvas</CODE>.
     *
     *    @return   The plot area background.
     *
    *********************************/
    public Color  getAreaBackground ()
    /********************************/
    {
	if (background == null)
	    return getBackground ();
	else
	    return background;
    }


    /**
     * Add the offset to the <I>X</I>&nbsp;data to the vector with
     * offsets of the plot traces. The default offset is zero.
     *
     *    @param  x_offset   The specified offset.
     *
    *****************************************/
    public void  addXOffset (Double  x_offset)
    /****************************************/
    {
	x_offsets.addElement (x_offset);
    }


    /**
     * Add the offset to the <I>Y</I>&nbsp;data to the vector with
     * offsets of the plot traces. The default offset is zero.
     *
     *    @param  x_offset   The specified offset.
     *
    *****************************************/
    public void  addYOffset (Double  y_offset)
    /****************************************/
    {
	y_offsets.addElement (y_offset);
    }


    /**
     * Set the flag which determines if the plot points should be joined
     * with line pieces as specified for the individual plot traces.
     *
     *    @param  use_join   <CODE>true</TT> if the points can be
     *			     joined; <CODE>false</CODE> otherwise.
     *
     *    @see  PlotCanvas#addJoin
     *
    ******************************************/
    public void  setUseJoin (boolean  use_join)
    /*****************************************/
    {
	this.use_join = use_join;
    }


    /**
     * Get the value of the flag which determines if the plot points
     * should be joined with line pieces.
     *
     *    @return   <CODE>true</TT> if the points can be joined.;
     *		    <CODE>false</CODE> otherwise.
     *
    ****************************/
    public boolean  getUseJoin ()
    /***************************/
    {
	return use_join;
    }


    /**
     * Add the flag which determines if the plot points should be joined
     * with line pieces to the vector with join flags for the plot
     * traces. The default is to join points.
     *
     *    @param  join_element   <CODE>true</TT> if the points of this
     *				 plot trace are joined;
     *				 <CODE>false</CODE> otherwise.
     *
     *    @see  PlotCanvas#setUseJoin
     *
    *******************************************/
    public void  addJoin (Boolean  join_element)
    /******************************************/
    {
	join.addElement (join_element);
    }


    /**
     * Remove all join points flags from the corresponding vector.
     *
     *    @see  PlotCanvas#removeJoin
     *    @see  PlotCanvas#removeJoinAt
     *
    *****************************/
    public void  removeAllJoins ()
    /****************************/
    {
	join.removeAllElements ();
    }


    /**
     * Remove the specified join points flag.
     *
     *    @param  join_element   The join flag component to remove.
     *
     *    @return  <CODE>true</CODE>if the argument was a component of
     *		   the join vector; <CODE>false</CODE> otherwise
     *
     *    @see  PlotCanvas#removeAllJoins
     *    @see  PlotCanvas#removeJoinAt
     *
    *************************************************/
    public boolean  removeJoin (Boolean  join_element)
    /************************************************/
    {
	return join.removeElement (join_element);
    }


    /**
     * Remove the specified join points flag at the specified index.
     *
     *    @param  index   The index of the join flag component to
     *			  remove.
     *
     *    @see  PlotCanvas#removeAllJoins
     *    @see  PlotCanvas#removeJoin
     *
    *************************************/
    public void  removeJoinAt (int  index)
    /************************************/
    {
	join.removeElementAt (index);
    }


    /**
     * Get the #&nbsp;join flags stored in the vector with join flags
     * for the plot traces.
     *
     *    @return   The #&nbsp;join flags.
     *
    ***********************/
    public int  getNrJoin ()
    /**********************/
    {
	return join.size ();
    }


    /**
     * Get the value of the join flag from the plot trace with the
     * specified index from the vector with join flags. When no flag is
     * found for the specified index <CODE>null</CODE> is returned.
     *
     *    @param  index   The index of the join flag to obtain.
     *
     *    @return   The join flag.
     *
    ***********************************/
    public Boolean  getJoin (int  index)
    /**********************************/
    {
	return (Boolean) join.elementAt (index);
    }


    /**
     * Set the flag which determines if markers should be used for the
     * plot points as specified for the individual plot traces.
     *
     *    @param  use_markers   <CODE>true</CODE> if the points can be
     *				marked; <CODE>false</CODE> otherwise.
     *
     *	  @see  PlotCanvas#addAllMarkers
     *    @see  PlotCanvas#addMarker(int)
     *    @see  PlotCanvas#addMarker(String)
     *    @see  PlotMarker
     *
    ************************************************/
    public void  setUseMarkers (boolean  use_markers)
    /***********************************************/
    {
	this.use_markers = use_markers;
    }


    /**
     * Get the flag which determines if markers should be used for the
     * plot points as specified for the individual plot traces.
     *
     *    @return   <CODE>true</CODE> if the points can be marked;
     *		    <CODE>false</CODE> otherwise.
     *
     *    @see  PlotMarker
     *
    *******************************/
    public boolean  getUseMarkers ()
    /******************************/
    {
	return use_markers;
    }

    /**
     * Add all markers to the vector with markers for the plot traces.
     * The default is to do not mark points.
     *
     *    @see  PlotCanvas#setUseMarkers
     *    @see  PlotCanvas#addMarker(int)
     *    @see  PlotCanvas#addMarker(String)
     *    @see  PlotMarker
     *
    ****************************/
    public void  addAllMarkers ()
    /***************************/
    {
	int []  markers    = PlotMarker.ALL_MARKERS;
	int	nr_markers = markers.length;

	for (int index = 0;  index < nr_markers;  index++)
	    addMarker (markers [index]);

    } // addAllMarkers ()


    /**
     * Add the specified marker to the vector with markers for the plot
     * traces. The default is to do not mark points.
     *
     *    @param  marker   The specified marker.
     *
     *    @see  PlotCanvas#setUseMarkers
     *	  @see  PlotCanvas#addAllMarkers
     *    @see  PlotCanvas#addMarker(String)
     *    @see  PlotMarker
     *
    ***********************************/
    public void  addMarker (int  marker)
    /**********************************/
    {
	markers.addElement (new PlotMarker (marker));
    }


    /**
     * Add the marker, specified with its name to the vector with
     * markers for the plot traces. The default is to do no mark points.
     *
     *    @param  marker_name   The specified marker name.
     *
     *    @see  PlotCanvas#setUseMarkers
     *	  @see  PlotCanvas#addAllMarkers
     *    @see  PlotCanvas#addMarker(int)
     *    @see  PlotMarker
     *
    *******************************************/
    public void  addMarker (String  marker_name)
    /******************************************/
    {
	markers.addElement (PlotMarker.valueOf (marker_name));
    }


    /**
     * Remove all markers from the vector with plot markers.
     *
     *    @see  PlotCanvas#removeMarker(int)
     *    @see  PlotCanvas#removeMarker(String)
     *    @see  PlotCanvas#removeMarkerAt
     *    @see  PlotMarker
     *
    *******************************/
    public void  removeAllMarkers ()
    /******************************/
    {
	markers.removeAllElements ();
    }


    /**
     * Remove the marker, specified by its value.
     *
     *    @param  marker_value   The marker value ID of the marker to
     *				 remove.
     *
     *    @return  <CODE>true</CODE> if the argument was a component of
     *		   the marker vector; <CODE>false</CODE> otherwise
     *
     *    @see  PlotCanvas#removeAllMarkers
     *    @see  PlotCanvas#removeMarker(String)
     *    @see  PlotCanvas#removeMarkerAt
     *    @see  PlotMarker
     *
    ***********************************************/
    public boolean  removeMarker (int  marker_value)
    /**********************************************/
    {
	int	 markers_size	= markers.size ();
	boolean  marker_removed = false;
	int	 index		= 0;

	/*
	 * Find the first marker specify by its value ID and remove it
	 * when found.
	 */

	while (index < markers_size  &&  !marker_removed)
	{
	    PlotMarker  marker = (PlotMarker) markers.elementAt (index);

	    if (marker.getValue () == marker_value)
		marker_removed = true;
	}

	return marker_removed;

    } // removeMarker ()


    /**
     * Remove the marker, specified by its name.
     *
     *    @param  marker_name   The name of the marker to remove.
     *
     *    @return  <CODE>true</CODE> if the argument was a component of
     *		   the marker vector; <CODE>false</CODE> otherwise
     *
     *    @see  PlotCanvas#removeAllMarkers
     *    @see  PlotCanvas#removeMarker(int)
     *    @see  PlotCanvas#removeMarkerAt
     *    @see  PlotMarker
     *
    *************************************************/
    public boolean  removeMarker (String  marker_name)
    /************************************************/
    {
	/*
	 * Parse the marker value ID from the specified name.
	 */

	int  marker_value = PlotMarker.getValueOf (marker_name);

	if (marker_value == -1)    // No such marker.
	{
	    return false;
	}
	else    // Remove the marker specified by its value ID.
	{
	    return removeMarker (marker_value);
	}

    } // removeMarker (String)


    /**
     * Remove the marker at the specified index.
     *
     *    @param  index   The index of the marker to remove.
     *
     *    @see  PlotCanvas#removeAllMarkers
     *    @see  PlotCanvas#removeMarker(int)
     *    @see  PlotCanvas#removeMarker(String)
     *    @see  PlotMarker
     *
    ***************************************/
    public void  removeMarkerAt (int  index)
    /**************************************/
    {
	markers.removeElementAt (index);
    }


    /**
     * Get the #&nbsp;markers stored in the vector with markers for the
     * plot traces.
     *
     *    @return   The #&nbsp;markers.
     *
     *    @see  PlotMarker
     *
    **************************/
    public int  getNrMarkers ()
    /*************************/
    {
	return markers.size ();
    }


    /**
     * Get the marker with the specified index from the vector with
     * markers for the plot traces. If no marker exists for the index
     * <CODE>null</CODE> is returned.
     *
     *    @param  index   The specified marker index.
     *
     *    @return   The marker of the plot trace with the specified
     *		    index.
     *
     *    @see  PlotMarker
     *
    ****************************************/
    public PlotMarker  getMarker (int  index)
    /***************************************/
    {
	return (PlotMarker) markers.elementAt (index);
    }


    /**
     * Set the minimum of the <I>X</I>&nbsp;values. If no minimum is
     * set, it will be determined from the plot data.
     *
     *    @param  xmin_obj   The specfied minimum.
     *
    **************************************/
    public void  setXMin (Double  xmin_obj)
    /*************************************/
    {
	this.xmin_obj = xmin_obj;
    }


    /**
     * Get the minimum of the <I>X</I>&nbsp;values. If no minimum is
     * set, <CODE>null</CODE> is returned.
     *
     *    @return  The minimum of the <I>X</I>&nbsp;values.
     *
    ************************/
    public Double  getXMin ()
    /***********************/
    {
	return xmin_obj;
    }


    /**
     * Set the maximum of the <I>X</I>&nbsp;values. If no maximum is
     * set, it will be determined from the plot data.
     *
     *    @param  xmax_obj   The specfied maximum.
     *
    **************************************/
    public void setXMax (Double  xmax_obj)
    /*************************************/
    {
	this.xmax_obj = xmax_obj;
    }


    /**
     * Get the maximum of the <I>X</I>&nbsp;values. If no maximum is
     * set, <CODE>null</CODE> is returned.
     *
     *    @return  The maximum of the <I>X</I>&nbsp;values.
     *
    ************************/
    public Double  getXMax ()
    /***********************/
    {
	return xmax_obj;
    }


    /**
     * Set the minimum of the <I>Y</I>&nbsp;values. If no minimum is
     * set, it will be determined from the plot data.
     *
     *    @param  ymin_obj   The specfied minimum.
     *
    **************************************/
    public void  setYMin (Double  ymin_obj)
    /*************************************/
    {
	this.ymin_obj = ymin_obj;
    }


    /**
     * Get the minimum of the <I>Y</I>&nbsp;values. If no minimum is
     * set, <CODE>null</CODE> is returned.
     *
     *    @return  The minimum of the <I>Y</I>&nbsp;values.
     *
    ************************/
    public Double  getYMin ()
    /***********************/
    {
	return ymin_obj;
    }


    /**
     * Set the maximum of the <I>Y</I>&nbsp;values. If no maximum is
     * set, it will be determined from the plot data.
     *
     *    @param  ymax_obj   The specfied maximum.
     *
    **************************************/
    public void  setYMax (Double  ymax_obj)
    /*************************************/
    {
	this.ymax_obj = ymax_obj;
    }


    /**
     * Get the maximum of the <I>Y</I>&nbsp;values. If no maximum is
     * set, <CODE>null</CODE> is returned.
     *
     *    @return  The maximum of the <I>Y</I>&nbsp;values.
     *
    ************************/
    public Double  getYMax ()
    /***********************/
    {
	return ymax_obj;
    }


    /**
     * Set the scaling type used at the <I>X</I>-axis. Linear or
     * logarithmic scaling can be used.
     *
     *    @param  x_scaling   The specified scaling type.
     *
     *    @see  PlotCanvas#LINEAR
     *    @see  PlotCanvas#LOG
     *
    ****************************************/
    public void  setXScaling (int  x_scaling)
    /***************************************/
    {
	this.x_scaling = x_scaling;
    }


    /**
     * Get the scaling type used at the <I>X</I>-axis. Linear or
     * logarithmic scaling can be used.
     *
     *    @return   The scaling type.
     *
     *    @see  PlotCanvas#LINEAR
     *    @see  PlotCanvas#LOG
     *
    *************************/
    public int  getXScaling ()
    /************************/
    {
	return x_scaling;
    }


    /**
     * Set the scaling type used at the <I>Y</I>-axis. Linear or
     * logarithmic scaling can be used.
     *
     *    @param  y_scaling   The specified scaling type.
     *
     *    @see  PlotCanvas#LINEAR
     *    @see  PlotCanvas#LOG
     *
    ****************************************/
    public void  setYScaling (int  y_scaling)
    /***************************************/
    {
	this.y_scaling = y_scaling;
    }


    /**
     * Get the scaling type used at the <I>Y</I>-axis. Linear or
     * logarithmic scaling can be used.
     *
     *    @return   The scaling type.
     *
     *    @see  PlotCanvas#LINEAR
     *    @see  PlotCanvas#LOG
     *
    *************************/
    public int  getYScaling ()
    /************************/
    {
	return y_scaling;
    }


    /**
     * Set the vector containing all plot data. Each plot trace is
     * represented by a vector which is stored in this vector.
     *
     *    @param  data   The specified vector with plot data.
     *
    ****************************************/
    public void  setDataVector (Vector  data)
    /***************************************/
    {
	this.data = data;
    }


    /**
     * Add a vector containing all data points of a plot trace to the
     * vector with plot traces. The data points are instances of the
     * <CODE>DataPoint</CODE> class. Positions where the joining of
     * lines should be interrupted can be marked with a
     * <CODE>null</CODE> <CODE>DataPoint</CODE> instance.
     *
     *    @param  trace   The specified vector with data points of the
     *			  plot trace.
     *
     *    @see  DataPoint
     *
    ****************************************/
    public void  addDataTrace (Vector  trace)
    /***************************************/
    {
	data.addElement (trace);
    }


    /**
     * Remove all plot traces from the vector with plot data.
     *
     *    @see  PlotCanvas#removeDataTrace
     *    @see  PlotCanvas#removeDataTraceAt
     *
    **********************************/
    public void  removeAllDataTraces ()
    /*********************************/
    {
	data.removeAllElements ();
    }


    /**
     * Remove the specified plot trace.
     *
     *    @param  trace   The plot trace to remove
     *
     *    @return  <CODE>true</CODE> if the argument was a component of
     *		   the vector with plot traces; <CODE>false</CODE>
     *		   otherwise
     *
     *    @see  PlotCanvas#removeAllDataTraces
     *    @see  PlotCanvas#removeDataTraceAt
     *
    **********************************************/
    public boolean  removeDataTrace (Vector  trace)
    /*********************************************/
    {
	return data.removeElement (trace);
    }


    /**
     * Remove the plot trace at the specified index.
     *
     *    @param  trace_index   The index of the plot trace to remove
     *
     *    @see  PlotCanvas#removeAllDataTraces
     *    @see  PlotCanvas#removeDataTrace
     *
    ************************************************/
    public void  removeDataTraceAt (int  trace_index)
    /***********************************************/
    {
	data.removeElementAt (trace_index);
    }


    /**
     * Return the #&nbsp;plot traces stored in the vector with plot
     * traces.
     *
     *    @return   The stored #&nbsp;plot traces.
     *
    *************************/
    public int  getDataSize ()
    /************************/
    {
	return data.size ();
    }


    /**
     * Get the vector with data points of the plot trace with the
     * specified number.
     *
     *    @param  trace_no   The specified index of the data plot trace.
     *
     *    @return   The data plot trace vector.
     *
    ********************************************/
    public Vector  getDataTraceAt (int  trace_no)
    /*******************************************/
    {
	return (Vector) data.elementAt (trace_no);
    }


    /**
     * Unzoom the plot domain to the full plot area.
     *
    *********************/
    public void  unZoom ()
    /********************/
    {
	zoomed = false;
    }


    /**
     * Initialise the plot canvas. Set also set some still undefined
     * parameters to their default values and set also the plot
     * variables connected with the data to plot.
     *
    ********************/
    private void  init ()
    /*******************/
    {
	canvas_size = size ();     // Set the current plot canvas size.

	/*
	 * Do nothing when at least one of the sizes is zero.
	 */

	if (canvas_size.width <= 0  ||  canvas_size.height <= 0)
	    return;

	if (background == null)		      // Set default background.
	    background = getBackground ();

	if (colors.size () == 0)		     // Set default
	    colors.addElement (getForeground ());    // trace color.

	if (x_offsets.size () == 0)		      // Set default
	    x_offsets.addElement (new Double (0.0));  // X trace offset.

	if (y_offsets.size () == 0)		      // Set default
	    y_offsets.addElement (new Double (0.0));  // Y trace offset.

	if (join.size () == 0)		       // Default join points.
	    join.addElement (Boolean.TRUE);

	if (markers.size () == 0)    // Default do not use markers.
	{
	    markers.addElement (new PlotMarker (PlotMarker.NONE));
	}

	/*
	 * Set the sizes of the plot area.
	 */
	 
	layoutArea ();

	/*
	 * Do nothing when at least one of the sizes of the plot area is
	 * zero.
	 */

	if (area.width <= 0  ||  area.height <= 0)
	    return;

	if (zoomed)    // Plot is zoomed in.
	{
	    /*
	     * Set the plot data dependent variables according to the
	     * data extrema in the zoom area.
	     */
	     
	    zoomScale ();
	}
	else    // The full domain is plotted.
	{
	    /*
	     * Scale the data. Take care of the scaling type and plot
	     * trace offsets.
	     */

	    scaleByData ();

	    /*
	     * Overrule the extrema with the values defined with the API
	     * methods.
	     */
	
	    if (xmin_obj != null)
		xmin = xmin_obj.doubleValue ();
	    
	    if (xmax_obj != null)
		xmax = xmax_obj.doubleValue ();

	    if (ymin_obj != null)
		ymin = ymin_obj.doubleValue ();
	    
	    if (ymax_obj != null)
		ymax = ymax_obj.doubleValue ();

	    /*
	     * Set a plot range when the minimum value is equal to the
	     * maximum value. Take linear logarithmic scaling into
	     * account.
	     */

	    if (xmin == xmax)
		if (x_scaling == LINEAR)
		{
		    xmin -= 0.5;
		    xmax += 0.5;
		}
		else    // LOG
		{
		    xmin /= 5.0;
		    xmax *= 5.0;
		}

	    if (ymin == ymax)
		if (y_scaling == LINEAR)
		{
		    ymin -= 0.5;
		    ymax += 0.5;
		}
		else    // LOG
		{
		    ymin /= 5.0;
		    ymax *= 5.0;
		}

	    /*
	     * Adjust the data extrema according to the plot ticks.
	     */

	    roundScaling ();

	    /*
	     * Again overrule the extrema with the values defined with
	     * the API methods.
	     */

	    if (xmin_obj != null)
		xmin = xmin_obj.doubleValue ();
	    
	    if (xmax_obj != null)
		xmax = xmax_obj.doubleValue ();
	    
	    if (ymin_obj != null)
		ymin = ymin_obj.doubleValue ();
	    
	    if (ymax_obj != null)
		ymax = ymax_obj.doubleValue ();

	    /*
	     * Calculate the tick positions using a.o. the rounded
	     * extrema.
	     */

	    calculateTickPositions ();
	}

    } // init ()


    /**
     * Resume the paint thread when it exists and it alive. This method
     * is typically intended to be called from the <CODE>start()</CODE>
     * method in the <CODE>Applet</CODE> class, when appropriate.
     *
     *    @see  PlotCanvas#suspendPaint
     *    @see  PlotCanvas#destroyPaint
     *    @see  PlotCanvas#quitPaint
     *
    **************************/
    public void  resumePaint ()
    /*************************/
    {
	if (paint_thread != null)    // Paint thread exists.
	{
	    if (paint_thread.isAlive ())    // Paint thread is running.
	    {
		paint_thread.resume ();    // Resume the paint thread.
	    }
	    else    // Paint thread is not running.
	    {
		paint_thread = null;    // Mark that it is finished.
	    }
	}

    } // resumePaint ()


    /**
     * Suspend the paint thread when it exists and it alive. This method
     * is typically intended to be called from the <CODE>stop()</CODE>
     * method in the <CODE>Applet</CODE> class, when appropriate.
     *
     *    @see  PlotCanvas#resumePaint
     *    @see  PlotCanvas#destroyPaint
     *    @see  PlotCanvas#quitPaint
     *
    ***************************/
    public void  suspendPaint ()
    /**************************/
    {
	if (paint_thread != null)    // Paint thread exists.
	{
	    if (paint_thread.isAlive ())    // Paint thread is running.
	    {
		paint_thread.suspend ();    // Suspend the paint thread.
	    }
	    else    // Paint thread is not running.
	    {
		paint_thread = null;    // Mark that it is finished.
	    }
	}

    } // suspendPaint ()


    /**
     * Destroy the paint thread when it exists and it alive. This method
     * is typically intended to be called from the
     * <CODE>destroy()</CODE> method in the <CODE>Applet</CODE> class,
     * when appropriate.
     *
     *    @see  PlotCanvas#resumePaint
     *    @see  PlotCanvas#suspendPaint
     *    @see  PlotCanvas#quitPaint
     *
    ***************************/
    public void  destroyPaint ()
    /**************************/
    {
	if (paint_thread != null)    // Paint thread exists.
	{
	    if (paint_thread.isAlive ())    // Paint thread is running.
	    {
		/*
		 * Mark the paint thread to quit and resume the paint
		 * thread which has been suspended by a previous call
		 * of the "suspendPaint()" method. This allows the paint
		 * thread to die.
		 */

		paint_quit.set	    ();    // Mark paint thread to quit.
		paint_thread.resume ();    // Resume the paint thread.
	    }
	    else    // Paint thread is not running.
	    {
		paint_thread = null;    // Mark that it is finished.
	    }
	}

    } // destroyPaint ()


    /**
     * Quit the painting thread and wait until it has been died. This
     * method is <B>not</B> connected with the
     * <CODE>resumePaint()</CODE>, <CODE>suspendPaint()</CODE> and
     * <CODE>destroyPaint()</CODE> methods which are typically intended
     * to be called from an <CODE>Applet</CODE> class.
     *
     *    @see  PlotCanvas#resumePaint
     *    @see  PlotCanvas#suspendPaint
     *    @see  PlotCanvas#destroyPaint
     *
    ************************/
    public void  quitPaint ()
    /***********************/
    {
	/*
	 * Quit the paint thread and wait until it dies when it is
	 * running and alive. Mark it.
	 */

	if (paint_thread != null)    // Paint thread exists.
	{
	    if (paint_thread.isAlive ())    // Paint thread is running.
	    {
		paint_quit.set ();    // Quit the paint thread.

		try    // Wait for the paint thread to quit.
		{
		    paint_thread.join ();
		}
		catch (InterruptedException  inter_except)
		{
		}
	    }
	    else    // Paint thread is not running.
	    {
		paint_thread = null;    // Mark that it is finished.
	    }
	}

    } // quitPaint ()


    /**
     * Update the plot. The plot canvas is completely initialised and
     * painted during the <CODE>paint()</CODE> function, so nothing has
     * to be done here.
     *
     *    @param  g   The graphics content for the update.
     *
     *    @see  PlotCanvas#paint
     *
    ********************************/
    public void  update (Graphics  g)
    /*******************************/
    {
	paint (g);    // Do all in this method.
    }


    /**
     * Paint the plot canvas. First initialise the plot canvas
     * parameters (sizes, and other plot parameters). It is always
     * completely painted. The actual painting is done in a private
     * thread which improves the performance of slow displays, because
     * the painting can be interrupted by a new <CODE>paint()</CODE>
     * call. This prevends a long stack of unhandled events at slow
     * displays.
     *
    *******************************/
    public void  paint (Graphics  g)
    /******************************/
    {
	/*
	 * Initalise the plot canvas.
	 */

	init ();

	/*
	 * Quit from a previous running paint thread and mark it.
	 */

	boolean  paint_stopped = false;

	if (paint_thread != null)    // Paint thread exists.
	{
	    if (paint_thread.isAlive ())    // Paint thread runs.
	    {
		paint_quit.set ();    // Mark the paint thread to quit.

		try    // Wait for the paint thread to quit.
		{
		    paint_thread.join ();
		}
		catch (InterruptedException  inter_except)
		{
		}

		paint_stopped = true;    // Mark thread has been quit.
	    }

	    paint_thread = null;    // Mark the thread non-existing.
	}

	/*
	 * Get a private graphics content instance for the paint thread.
	 */

	if (paint_stopped)    // A running paint thread has been quit.
	{
	    /*
	     * Create a new graphics content for the plot canvas. Set
	     * the clip rectangle to the full canvas, because when do
	     * not know what was done by the interrupted paint, so the
	     * savest thing is to completely repaint the plot.
	     */

	    paint_graphics = getGraphics ();
	    paint_graphics.clipRect (0, 0, canvas_size.width,
				     canvas_size.height);
	}
	else    // No paint thread was running.
	{
	    /*
	     * Create a copy of the graphical contect, specified to this
	     * method.
	     */

	    paint_graphics = g.create ();
	}

	/*
	 * The paint thread, created below, shoud start to run.
	 */

	paint_quit.reset ();

	/*
	 * Create a new paint thread and start the painting of the plot.
	 */

	paint_thread = new Thread (this);
	paint_thread.start ();

    } // paint ()


    /**
     * Paint the plot canvas in the paint thread. Jump from this method
     * when a <CODE>QuitException</CODE> has been thrown by the paint
     * thread. It will be thrown when the thread should stop.
     *
    ******************/
    public void  run ()
    /*****************/
    {
	/*
	 * Catch the Quit Exception which will be thrown when the blink
	 * thread should stop.
	 */

	try
	{
	    paintPlot (paint_graphics);    // Paint the plot canvas.
	}
	catch (QuitException  quit_except)
	{
	}

	paint_graphics = null;    // Graphical context no more needed.
    }


    /**
     * Paint the plot canvas. The canvas will be painted
     * completely. A <CODE>QuitException</CODE> will be thrown when the
     * paint thread is requested to quit.
     *
     *    @param  g   The graphical context used to paint the plot.
     *
     *    @exception  QuitException   Thrown when the paint thread
     *				      should quit.
     *
    ***********************************************************/
    private void  paintPlot (Graphics  g)  throws  QuitException
    /**********************************************************/
    {
	/*
	 * When one of the plot area sizes is zero, fill the
	 * non-zero plot canvas with the default background and stop.
	 */

	if (area.width <= 0  ||  area.height <= 0)
	{
	    fillCanvasBackground (g);
	    return;
	}

	/*
	 * Paint the plot canvas.
	 */

	clearCanvas (g); // Clear canvas and plot area with the bckgd's.
       	drawAxes    (g); // Draw axes with labels, ticks and tick lab's.
	plotData    (g); // Plot the data in the traces.

    } // paintPlot ()


    /**
     * Handle the event when a mouse entered the plot canvas. Show the
     * corresponding status message when the mouse did enter the plot
     * canvas. In all other situation the super method is called to
     * handle the event.
     * </P>
     *
     * <P>
     * The function returns <CODE>true</CODE> when the event was
     * handled; <CODE>false</CODE> otherwise.
     *
     *    @param  evt   The event that caused the action.
     *
     *    @param  x	The <I>X</I>&nbsp;position of the mouse pointer.
     *
     *    @param  y	The <I>Y</I>&nbsp;position of the mouse pointer.
     *
     *    @return   <CODE>true</CODE> when the event was handled;
     *		    <CODE>false</CODE> otherwise.
     *
    **************************************/
    public boolean  mouseEnter (Event  evt,
				int    x,
				int    y)
    /*************************************/
    {
	if (evt.target == this)    // Target is the plot canvas.
	{
	    /*
	     * Let the "PlotStatusHandler" handler interface show the
	     * status message. Inform the interface that the mouse
	     * entered the canvas.
	     */

	    status_handler.showMessage ("Use mouse to zoom in.", ENTER);

	    return true;
	}
	else    // Another target: handled by the super method.
	{
	    return super.mouseEnter (evt, x, y);
	}

    } // mouseEnter ()


    /**
     * Handle the event when a mouse exited the plot canvas. Clear the
     * corresponding status message when the mouse did exit the plot
     * canvas. In all other situation the super method is called to
     * handle the event.
     * </P>
     *
     * <P>
     * The function returns <CODE>true</CODE> when the event was
     * handled; <CODE>false</CODE> otherwise.
     *
     *    @param  evt   The event that caused the action.
     *
     *    @param  x	The <I>X</I>&nbsp;position of the mouse pointer.
     *
     *    @param  y	The <I>Y</I>&nbsp;position of the mouse pointer.
     *
     *    @return   <CODE>true</CODE> when the event was handled;
     *		    <CODE>false</CODE> otherwise.
     *
    ************************************/
    public boolean  mouseExit (Event evt,
    			       int   x,
			       int   y)
    /***********************************/
    {
	if (evt.target == this)    // Target is the plot canvas.
	{
	    /*
	     * Let the "PlotStatusHandler" handler interface clear the
	     * status message. Inform the interface that the mouse
	     * exited the canvas.
	     */
	     
	    status_handler.showMessage ("", EXIT);

	    return true;
	}
	else    // Another target: handled by the super method.
	{
	    return super.mouseExit (evt, x, y);
	}

    } // mouseExit ()
    
    
    /** Handle the event when the user has clicked in the plot canvas.
     * The zoom area is marked when the left mouse button is clicked.
     * </P>
     *
     * <P>
     * The function returns <CODE>true</CODE> when the event was
     * handled; <CODE>false</CODE> otherwise.
     *
     *    @param  evt	The event that caused the action.
     *
     *    @param  x	The <I>X</I>&nbsp;position of the mouse pointer.
     *
     *    @param  y	The <I>Y</I>&nbsp;position of the mouse pointer.
     *
     *    @return   <CODE>true</CODE> when the event was handled;
     *		    <CODE>false</CODE> otherwise.
     *
    *************************************/
    public boolean  mouseDown (Event  evt,
			       int    x,
			       int    y)
    /************************************/
    {
	/*
	 * Let the super method handle the mouse click when the target
	 * was not the plot canvas.
	 */

	if (evt.target != this)
	    return super.mouseDown (evt, x, y);

	/*
	 * Do nothing when:
	 *
	 *    1. One of the canvas sizes was zero.
	 *
	 *    2. Not the first mouse button was clicked. This implies
	 *	 that the META or ALT mask bits are non-zero.
	 */

	if (canvas_size.width == 0  ||  canvas_size.height == 0  ||
	    (evt.modifiers & (Event.META_MASK | Event.ALT_MASK)) != 0)
	{
	    return true;
	}

	/*
	 * Mark the point and notify that the mouse is not yet dragged.
	 */
	    
	mouse_down = new Point (x,y);
	mouse_drag = false;
	
	return true;    // The mouse click has been handled.

    } // mouseDown ()

    
    /**
     * Handle the event when the mouse was dragged in the plot canvas.
     * The zoom area selection is started when the mouse was not yet
     * dragged; it is extended when the mouse was alreday dragged.
     * handled; <CODE>false</CODE> otherwise.
     * </P>
     *
     * <P>
     * The function returns <CODE>true</CODE> when the event was
     * handled; <CODE>false</CODE> otherwise.
     *
     *    @param  evt	The event that caused the action.
     *
     *    @param  x	The <I>X</I>&nbsp;position of the mouse pointer.
     *
     *    @param  y	The <I>Y</I>&nbsp;position of the mouse pointer.
     *
     *    @return   <CODE>true</CODE> when the event was handled;
     *		    <CODE>false</CODE> otherwise.
     *
    *************************************/
    public boolean  mouseDrag (Event  evt,
			       int    x,
			       int    y)
    /************************************/
    {
        /*
         * Let the super method handle the mouse click when the target
	 * was not the plot canvas.
	 */

	if (evt.target != this)
	    return super.mouseDrag (evt, x, y);

	/*
	 * Do nothing when:
         *
         *    1. One of the canvas sizes was zero.
         *
         *    2. Not the first mouse button was clicked. This implies
         *       that the META or ALT mask bits are non-zero.
         */

	if (canvas_size.width == 0  ||  canvas_size.height == 0  ||
	    (evt.modifiers & (Event.META_MASK | Event.ALT_MASK)) != 0)
	{
	    return true;
	}

	/*
	 * Store the current value of the mouse drag variable. It will
	 * be compared below to check if the dragging was started.
	 */

	boolean  already_mouse_drag = mouse_drag;

	/*
	 * Require a minimum start zoom length before a mouse movement
	 * is marked for the first time as a mouse drag.
	 */

	if (Math.abs (mouse_down.x) + Math.abs (mouse_down.y) >
	    ZOOM_LENGTH_BOUND)
	{
	    mouse_drag = true;
	}

	/*
	 * Start the zoom area selection when the mouse was dragged for
	 * the first time.
	 */
	  
	if (mouse_drag && ! already_mouse_drag)
	    startSelection (x, y);

	/*
	 * Extend the zoom area selection when the mouse dragging
	 * continues.
	 */

	if (mouse_drag)
	    extendSelection (x, y);

	return true;    // A mouse drag event has been handled.

    } // mouseDrag ()

    
    /**
     * Handle the event when the mouse button went up.
     *
     * <UL>
     *
     * <LI>If the left button went up, we zoom in to the zoom area.</LI>
     *
     * <LI>If the right button went up, we zoom out to the full plot
     *	   area.</LI>
     *
     * </UL>
     * </P>
     *
     * <P>
     * The function returns <CODE>true</CODE> when the event was
     * handled; <CODE>false</CODE> otherwise.
     *
     *    @param  evt	The event that caused the action.
     *
     *    @param  x	The <I>X</I>&nbsp;position of the mouse pointer.
     *
     *    @param  y	The <I>Y</I>&nbsp;position of the mouse pointer.
     *
     *    @return   <CODE>true</CODE> when the event was handled;
     *		    <CODE>false</CODE> otherwise.
     *
    ***********************************/
    public boolean  mouseUp (Event  evt,
			     int    x, 
			     int    y)
    /**********************************/
    {
	/*
	 * Let the super method handle the event when the plot canvas
	 * was not the target.
	 */

	if (evt.target != this)
	    return super.mouseUp (evt, x, y);

	/*
	 * Do nothing when one of the plot canvas sizes was zero.
	 */

	if (canvas_size.width == 0  ||  canvas_size.height == 0)
	    return true;

	/*
	 * Handle the button up event.
	 */
	 
	if ((evt.modifiers & (Event.META_MASK | Event.ALT_MASK)) != 0)
	{
	    /*
	     * Right button went up: zoom out to the full plot area,
	     * repaint the plot canvas.
	     */

	    autoScale ();
	    repaint   ();
	}
	else    // Left button went up.
	{ 
	    /*
	     * Only end the selection of the zoom area and zoom in when
	     * the start "length" of the zoom area exceeds a boundary;
	     * repaint then also the plot cavas.
	     */

	    if (selection != null  &&
	        Math.abs (selection.x) + Math.abs (selection.y) >
		ZOOM_LENGTH_BOUND)
	    {
	        endSelection (x, y);    // Draw last zoom area.

		zoomIn  ();
		repaint ();

		selection = null;    // Clear zoom selection.
	    }
	}

	return true;    // A mouse up event was handled.

    } // mouseUp ()

    
    /**
     * Start the selection of the zoom area. The first zoom area is
     * painted in XOR mode in the plot canvas.
     *
     *    @param  x   The <I>X</I>-coordinate of the south-east corner
     *		      of the selection rectangle.
     *
     *    @param  y   The <I>Y</I>-coordinate of the south-east corner
     *		      of the selection rectangle.
     *
    ************************************/
    private void  startSelection (int  x,
				  int  y)
    /***********************************/
    {
	/*
	 * Create a new selection rectangle.
	 */

	selection = new Rectangle (mouse_down.x,     mouse_down.y,
				   x - mouse_down.x, y - mouse_down.y);

	/*
	 * If height or width were negative the rectangle would not be
	 * displayed properly.
	 */

	if (selection.width < 0)
	{
	    selection.x     += selection.width;
	    selection.width *= -1;
	}

	if (selection.height < 0)
	{
	    selection.y      += selection.height;
	    selection.height *= -1;
	}

	/*
	 * Draw the new selection rectangle XOR'd with the plot canvas.
	 */

	Graphics  g = getGraphics ();

	g.setColor   (background);
	g.setXORMode (getForeground ());
	g.drawRect   (selection.x,     selection.y,
		      selection.width, selection.height);

    } // startSelection ()

    
    /**
     * Extend the selection of the zoom area. The previous zoom area is
     * cleared and the extended one is painted; both in XOR mode.
     *
     *    @param  x   The <I>X</I>-coordinate of the south-east corner
     *		      of the selection rectangle.
     *
     *    @param  y   The <I>Y</I>-coordinate of the south-east corner
     *		      of the selection rectangle.
     *
    *************************************/
    private void  extendSelection (int  x,
				   int  y)
    /************************************/
    {
	Graphics  g = getGraphics ();

	/*
	 * Clear previous selection rectangle.
	 */

	g.setColor   (background);
	g.setXORMode (getForeground ());
	  
	if (selection != null) 
	    g.drawRect (selection.x,     selection.y,
			selection.width, selection.height);

	selection = new Rectangle (mouse_down.x,     mouse_down.y,
				   x - mouse_down.x, y - mouse_down.y);

	/*
	 * If height or width were negative the rectangle would not be
	 * displayed properly.
	 */

	if (selection.width < 0)
	{
	    selection.x     += selection.width;
	    selection.width *= -1;
	}
	
	if (selection.height < 0)
	{
	    selection.y      += selection.height;
	    selection.height *= -1;
	}

	/*
	 * Paint the extended selection area in XOR mode.
	 */
	
	g.drawRect (selection.x,     selection.y,
		    selection.width, selection.height);

    } // extendSelection ()

    
    /**
     * End the selection of the zoom area. The previous zoom area is
     * cleared and the final one is painted; both in XOR mode.
     *
     *    @param  x   The <I>X</I>-coordinate of the south-east corner
     *		      of the selection rectangle.
     *
     *    @param  y   The <I>Y</I>-coordinate of the south-east corner
     *		      of the selection rectangle.
     *
    **********************************/
    private void  endSelection (int  x,
				int  y)
    /*********************************/
    {
	Graphics  g = getGraphics ();

	/*
	 * Clear previous selection rectangle.
	 */
	   
	g.setColor   (background);
	g.setXORMode (getForeground());

	if (selection != null) 
	    g.drawRect (selection.x,     selection.y,
			selection.width, selection.height);

	selection = new Rectangle (mouse_down.x,     mouse_down.y,
				   x - mouse_down.x, y - mouse_down.y);

	/*
	 * If height or width were negative the rectangle would not be
	 * displayed properly.
	 */

	if (selection.width < 0)
	{
	    selection.x     += selection.width;
	    selection.width *= -1;
	}

	if (selection.height < 0)
	{
	    selection.y      += selection.height;
	    selection.height *= -1;
	}

	/*
	 * Draw final selection rectangle. No XORMode.
	 */
	 
	g.setPaintMode ();
	g.setColor     (getForeground());
	g.drawRect     (selection.       x,selection.y,
			selection.width, selection.height);

    } // endSelection ()


    /**
     * Layout the plot area rectangle.
     *
    **************************/
    private void  layoutArea ()
    /*************************/
    {
	/*
	 * Initialise the plot area rectangle with equal sizes as the
	 * plot canvas.
	 */

	area = new Rectangle (canvas_size);

	/*
	 * Set a border around the plot area.
	 */
	
	area.x	    += OUTER_AREA_BORDER;
	area.width  -= 2 * OUTER_AREA_BORDER; 
	area.y	    += OUTER_AREA_BORDER;
	area.height -= 2 * OUTER_AREA_BORDER;
	
	/*
	 * Inquire the font metrics to reserve space for titles and
	 * labels.
	 */

	FontMetrics  font = getFontMetrics (getFont ());
	
	int  font_height  = font.getHeight   ();
	int  number_width = font.stringWidth ("0");
	// int  char_width  = font.getMaxAdvance ();

	/*
	 * Leave space for x scaling marks.
	 */

	area.height -= font_height * x_handler.getNrTickLines ();

	/*
	 * Leave space for y scaling marks.
	 */

	area.y += font_height / 2;

	area.height -= font_height * y_handler.getNrTickLines ();


	if (x_axis_label != null)
	    area.height -= font_height;  // leave space for x axis label

	if (y_axis_label != null)    // leave space for y axis label
	{
	    area.y	+= font_height;  // / 2;
	    area.height -= font_height;
	}

        area.x     += 7 * number_width;
	area.width -= 7 * number_width;  // space for y scaling marks 
	area.width -= 3 * number_width;  // space for x scaling marks
    
    } // layoutArea ()

    
    /**
     * Zoom in to the selected zoom area.
     *
    **********************/
    private void  zoomIn ()
    /*********************/
    {
	/*
	 * Do nothing when the selection does not exist or when one of
	 * the sizes is zero.
	 */

	if (selection == null ||
	    selection.width == 0 || selection.height == 0)
	{
	    return;
	}

	/*
	 * Let the PlotStatusHandler interface show the message that we
	 * are zooming in.
	 */

	status_handler.showMessage ("Zooming in...", WITHIN);
	
	/*
	 * Translate the corners of the zoom selection rectangle to
	 * the imzoomed data point extrema.
	 */

	xmin_zoom = xValue (selection.x);
	xmax_zoom = xValue (selection.x + selection.width); 
	ymax_zoom = yValue (selection.y);
	ymin_zoom = yValue (selection.y + selection.height);

	zoomed = true;    // Notify that we are zoomed in.

	/*
	 * Set the plot data dependent variables according to the data
	 * extrema in the zoom area.
	 */
	     
	zoomScale ();

	/*
	 * Let the PlotStatusHandler interface show the message how to
	 * zoom out.
	 */

	status_handler.showMessage (
	    "Alt+Click or Right-Mouse to zoom out.", WITHIN);

    } // zoomIn ()


    /**
     * Set the plot data dependent variables according to the data
     * extrema in the zoom area.
     *
    *************************/
    private void  zoomScale ()
    /************************/
    {
	/*
	 * Do nothing when one of the plot area sizes is zero.
	 */

	if (area.width <= 0  ||  area.height <= 0)
	    return;

	/*
	 * Set the data extrema to the imzoomed extrema.
	 */

	xmin = xmin_zoom;
	xmax = xmax_zoom;
	ymin = ymin_zoom;
	ymax = ymax_zoom; 
	  
	/*
	 * Adjust the data extrema according to the plot ticks and
	 * calculate the plot ticks.
	 */

	roundScaling	       ();
	calculateTickPositions ();

    } // zoomScale ()

    
    /**
     * Zoom out and rescale to the full plot canvas.
     *
    *************************/
    private void  autoScale ()
    /************************/
    {
	zoomed = false;    // Notify that we are zoomed out.

	/*
	 * Let the "PlotStatusHandler" interface show the message that
	 * we are autoscaling.
	 */

	status_handler.showMessage ("Autoscale...", WITHIN);

	/*
	 * Rescale the plot data in the full domain.
	 */

	rescale ();

    } // autoScale ()


    /**
     * Rescale the data and calculate the tick positions in the full
     * plot domain.
     *
    ***********************/
    private void  rescale ()
    /**********************/
    {
	/*
	 * Do nothing when one af the plot area sizes is zero.
	 */

	if (area.width <= 0  ||  area.height <= 0)
	    return;

	/*
	 * Scale the data. Take care of the scaling type and plot trace
	 * offsets.
	 */

	scaleByData ();
	  
	/*
	 * Set a plot range when the minimum value is equal to the
	 * maximum value.
	 */

	if (xmin == xmax)
	{
	    xmin -= 0.5;
	    xmax += 0.5;
	}

	if (ymin == ymax)
	{
	    ymin -= 0.5;
	    ymax += 0.5;
	}
	  
	/*
	 * Adjust the data extrema according to the plot ticks.
	 */

	roundScaling ();

	/*
	 * Calculate the tick positions using a.o. the rounded extrema.
	 */

	calculateTickPositions ();

    } // rescale ()

    
    /**
     * Scale the data. Calculate the data extrema. Take care of the plot
     * trace offsets.
     *
    ***************************/
    private void  scaleByData ()
    /**************************/
    {
	/*
	 * Set the X and Y coordinate extrema to the first valid scaled
	 * coordinates found in the plot traces.
	 */

	first:
	    for (int i = 0;  i < data.size ();  i++)
	    {
		Vector trace = (Vector) data.elementAt (i);

		if (trace.size () == 0)    // Go to the next trace when
		    continue;		   // it does not contain data.
	    
		/*
		 * Obtain the X and Y offsets for this plot trace. They
		 * are obtained by the trace index modulo the # traces.
		 */

		double  xo = ( (Double) x_offsets.elementAt (
					    i % x_offsets.size())
			     ).doubleValue ();
		double  yo = ( (Double) y_offsets.elementAt (
					    i % y_offsets.size())
			     ).doubleValue ();

		/*
		 * Do for all points in the trace UNTIL THE FIRST VALID
		 * POINT IS FOUND.
		 */

		for (int j = 0;  j < trace.size ();  j++)
		{
		    /*
		     * Get a new data point with the specified index.
		     */

		    DataPoint  p = (DataPoint) trace.elementAt (j);

		    /*
		     * Go to the next point when it is null.
		     */

		    if (p == null)
			continue;

		    /*
		     * Get the coordinates of a new data point.
		     */

		    double  x = p.x;
		    double  y = p.y;

		    /*
		     * Correct for the offsets. Scale and scale back to
		     * check if the point is valid.
		     */

		    x = invXScale (xScale (x + xo));
		    y = invYScale (yScale (y + yo)); 

		    /*
		     * Go to the next point when one of the coordinates
		     * is Not a Number.
		     */

		    if (Double.isNaN(x) || Double.isNaN(y))
			continue;

		    /*
		     * Set the extrema to the first valid point.
		     */

		    xmin = xmax = x;
		    ymin = ymax = y;
		    
		    break first;    // BREAK OUT THE 'first' LABEL AFTER
		    		    // THE FIRST VALID DATA POINT.

		} // For all data in a plot trace.

	    } // For all plot traces.

	// END OF THE 'first' LABEL.

	/*
	 * Do for all plot traces.
	 */

	for (int i = 0;  i < data.size ();  i++)
	{
	    Vector  trace = (Vector) data.elementAt (i);
	    
	    double  xo = ( (Double) x_offsets.elementAt (
				        i % x_offsets.size())
			 ).doubleValue ();
	    double  yo = ( (Double) y_offsets.elementAt (
					i % y_offsets.size())
			 ).doubleValue ();
        
	    /*
	     * Do for all data points in a plot trace.
	     */

	    for (int j = 0;  j < trace.size ();  j++)
	    {
		/*
		 * Get a new data point with the specified index.
		 */

		DataPoint  p = (DataPoint) trace.elementAt (j);

		/*
		 * Go to the next data point when it is null.
		 */

		if (p == null)
		    continue;

		double  x = p.x;
		double  y = p.y;
		
		x = invXScale (xScale (x + xo));
		y = invYScale (yScale (y + yo));

		/*
		 * Set a new X extremum when found.
		 */

		if (x < xmin)
		    xmin = x;
		else if (x > xmax)
		    xmax = x;

		/*
		 * Set a new Y extremum when found.
		 */

		if (y < ymin)
		    ymin = y;
		else if (y > ymax)
		    ymax = y;

	    } // For all data points in a plot trace.

	} // For all plot traces.

    } // scaleByData ()

	
    /**
     * Round all data points to even numbers (when possible).
     * Used by "zoomIn" and "autoScale".
     * Plot looks better if x and y axes start and end with an even
     * number (round to even number only when the implementation of the
     * <CODE>DataHandler</CODE> from this class is used.
     *
     *    @see DataHandler#round
     *
    ***************************/
    private void roundScaling ()
    /**************************/
    {
	/*
	 * Set the respective coordinates which are one tick space
	 * more (less) than the maximum (minimum).
	 */

	double  x1 = xValue (xPixel (xmin) - x_tick_spacing); 
	double  x2 = xValue (xPixel (xmax) + x_tick_spacing);
	double  y1 = yValue (yPixel (ymin) + y_tick_spacing);
	double  y2 = yValue (yPixel (ymax) - y_tick_spacing);

	/*
	 * Round these coordinates, using the "DataHandler" interface.
	 * Differences with the original stay with a tick space
	 * distance.
	 */

	xmin = x_handler.round (xmin, x1 - xmin);
	xmax = x_handler.round (xmax, x2 - xmax);
	ymin = y_handler.round (ymin, y1 - ymin);
	ymax = y_handler.round (ymax, y2 - ymax);

    } // roundScaling ()


    /**
     * Calculate the tick postions from the plot area rectangle
     * coordinates and the data extrema.
     *
    **************************************/
    private void  calculateTickPositions ()
    /*************************************/
    {
	/*
	 * Calculate the X-tick positions.
	 */

	x_ticks.setSize (0);    // Clear the X-ticks vector.

	if (area.width <= 0)    // Stop when the X-dimension is zero.
	    return;

	if (x_scaling == LOG)    // Logarithmic X-axis scaling.
	{
	    /*
	     * Initialise some variables.
	     */

	    double   x1		= 0.0;      // The compiler want these
	    boolean  round_up	= false;    // variables initialised.
	    boolean  init_ticks = false;    // Ticks must be initialsed.
	    boolean  cont_ticks = true;     // Continue adding ticks.

	    /*
	     * Add ticks to the vector with X-ticks.
	     */

	    while (cont_ticks)
	    {
		double  x2;
		double  x_pow;
		double  x_dec;

		if (!init_ticks)    // Initialise the ticks.
		{
		    /*
		     * Convert the first X-pixel of the plot cavas to a
		     * X-value. Do the same for the X-pixel at the
		     * specified tick distance.
		     */

		    x1 = xValue (area.x);
		    x2 = xValue (area.x + x_tick_spacing);

		    /*
		     * If the specified X-values are decreasing
		     * (increasing) the rounding should be up (down):
		     * the specified tick spacing is the maximum
		     * spacing.
		     */

		    round_up = (x1 < x2 ? false : true);

		    /*
		     * Calculate the Y-decade nearest to "x1" when going
		     * to "x2".
		     */
		     
		    if (round_up)
			x_pow = Math.floor (log10 (x1));
		    else
			x_pow = Math.ceil  (log10 (x1));

		    x_dec = Math.pow (10.0, x_pow);

		    /*
		     * Use the found X-decade as the first tick if it is
		     * situated between "x1" and "x2". This gives a
		     * nicer tick distribution.
		     *
		     * When "x2" is larger than the first X-decade, the
		     * first tick is rounded to a value as nice as
		     * possible situated between "x1" and "x2".
		     */

		    if (  round_up  &&  x_dec > x2  ||  // Use X-decade.
			 !round_up  &&  x_dec < x2 )
		    {
			x1 = x_dec;
		    }
		    else    // Round between "x1" and "x2".
		    {
			x1 = x_handler.round (x1, x2 - x1);
		    }

		    init_ticks = true;    // The ticks are initialised.

		} // Initialise the ticks.

		/*
		 * Add the value of the X-tick.
		 */

		x_ticks.addElement (new Double (x1));

		/*
		 * Get the next X-value at the maximum, specified tick
		 * spacing.
		 */

		x2 = xValue (xPixel (x1) + x_tick_spacing);

		/*
		 * Get the next decade of this X-value ("x2") in the
		 * direction of the rounding orientation.
		 */

		if (round_up)
		    x_pow = Math.ceil  (log10 (x2));
		else
		    x_pow = Math.floor (log10 (x2));

		x_dec = Math.pow (10.0, x_pow);

		/*
		 * Use the decade as next tick X-value when it is
		 * smaller than the current tick X-value ("x1");
		 * otherwise, do a rounding.
		 */

		if (round_up && x_dec < x1  ||  !round_up && x_dec > x1)
		    x1 = x_dec;
		else
		    x1 = x_handler.round (x2, x1 - x2);

		/*
		 * Stop with the addition of ticks when the boundary of
		 * the plot canvas is crossed.
		 */

		if (xPixel (x1) > area.x + area.width)
		    cont_ticks = false;

	    } // Add ticks to the vector with X-ticks.
	}
	else    // Linear X-axis scaling.
	{
	    /*
	     * Set "x1" to the first value in the plot area and set "x2"
	     * to the value one tick further (in the area domain).
	     */

	    double  x1 = xValue (area.x);
	    double  x2 = xValue (area.x + x_tick_spacing);

	    /*
	     * Round "x1" to the next value. Because the extrema are
	     * rounded with an absolute larger range, this value will
	     * never be larger than the first minimum.
	     */

	    x1 = x_handler.round (x1, x2 - x1);

	    /*
	     * Round "x2" to the previous value > "x1".
	     */
	
	    x2 = xValue		  (xPixel (x1) + x_tick_spacing);
	    x2 = x_handler.round  (x2, x1 - x2);

	    /*
	     * The difference between "x2" and "x1" is the tick step in
	     * the X-coordinate domain; "x1" is the first tick step
	     * (=> "x").
	     */

	    double  x_step = x2 - x1;
	    double  x      = x1;

	    /*
	     * Set the X-ticks while whe are within or at the border of
	     * the plot area.
	     */

	    while (xPixel (x) <= area.x + area.width)
	    {
		x_ticks.addElement (new Double (x));  // Add the X-tick.

		/*
		 * Get the next X-tick from the DataHandler: the default
		 * implementation in this class is:
		 *
		 *		"x += x_step"
		 */

		x = x_handler.nextTickValue (x, x_step);
	    }

	    /*
	     * Set the Y-ticks in a comparable way as the X-ticks.
	     */
      
	    y_ticks.setSize (0);

	} // Linear X-axis scaling.

	/*
	 * Set the Y-ticks in a comparable way as the X-ticks.
	 */

	if (area.height <= 0)
	    return;

	if (y_scaling == LOG)    // // Logarithmic Y-axis scaling.
	{
	    double   y1		= 0.0;
	    boolean  round_up	= false;
	    boolean  init_ticks = false;
	    boolean  cont_ticks = true;

	    while (cont_ticks)
	    {
		double  y2;
		double  y_pow;
		double  y_dec;

		if (!init_ticks)
		{
		    y1 = yValue	(area.y);
		    y2 = yValue	(area.y + y_tick_spacing);

		    round_up = (y1 < y2 ? false : true);

		    if (round_up)
			y_pow = Math.floor (log10 (y1));
		    else
			y_pow = Math.ceil  (log10 (y1));

		    y_dec = Math.pow (10.0, y_pow);

		    if (  round_up  &&  y_dec > y2  ||
			 !round_up  &&  y_dec < y2 )
		    {
			y1 = y_dec;
		    }
		    else
		    {
			y1 = y_handler.round (y1, y2 - y1);
		    }

		    init_ticks = true;
		}

		y_ticks.addElement (new Double (y1));

		y2 = yValue (yPixel (y1) +  y_tick_spacing);

		if (round_up)
		    y_pow = Math.ceil  (log10 (y2));
		else
		    y_pow = Math.floor (log10 (y2));

		y_dec = Math.pow (10.0, y_pow);

		if (round_up && y_dec < y1  ||  !round_up && y_dec > y1)
		    y1 = y_dec;
		else
		    y1 = y_handler.round (y2, y1 - y2);

		if (yPixel (y1) > area.y + area.height)
		    cont_ticks = false;
	    }
	}
	else    // Linear Y-axis scaling.
	{
	    double  y1 = yValue (area.y);
	    double  y2 = yValue (area.y + y_tick_spacing);

	    y1 = y_handler.round (y1, y2 - y1);

	    y2 = yValue	         (yPixel (y1) + y_tick_spacing);
	    y2 = y_handler.round (y2, y1 - y2);

	    double  y_step = y2 - y1;
	    double  y	   = y1;

	    while (yPixel (y) <= area.y + area.height)
	    {
		y_ticks.addElement (new Double (y));

		y = y_handler.nextTickValue (y, y_step);
	    }  

	} // Linear Y-axis scaling.

    } // calculateTickPositions ()


    /**
     * Fill the plot canvas with the background color. Check first if
     * the paint thread should quit. In that case a
     * <CODE>QuitException</CODE> will be thrown.
     *
     *    @param  g   The graphical context to use.
     *
     *    @exception  QuitExcpetion   Thrown when the paint thread
     *				      should quit.
     *
    *******************************************************/
    private void  fillCanvasBackground ( Graphics  g
				       )
				       throws  QuitException
    /******************************************************/
    {
	paint_quit.checkThrow ();   // Check quit of paint thread.

	g.setPaintMode ();
	g.setColor     (getBackground ());
	g.fillRect     (0, 0, canvas_size.width, canvas_size.height);
    }

    
    /**
     * Clear the plot canvas. The part of the plot canvas around the
     * plot area is filled with the canvas background color. The plot
     * area within the plot canvas is filled with the area background.
     * Check first if the paint thread should quit. In that case a
     * <CODE>QuitException</CODE> will be thrown.
     *
     *    @param  g   The graphical context to use.
     *
     *    @exception  QuitException   Thrown when the paint thread
     *				      should quit.
     *
    *************************************************************/
    private void  clearCanvas (Graphics  g)  throws  QuitException
    /************************************************************/
    {
	paint_quit.checkThrow ();    // Check quit of paint thread.

	g.setPaintMode ();

	/*
	 * Fill the part of the plot canvas around the plot area with
	 * the canvas background color, but do nothing with the plot
	 * area itsself to reduce flicker. Therefore the filling is
	 * split into four rectangles. 
	 */

	g.setColor (getBackground ());
	g.fillRect (0, 0, canvas_size.width, area.y);
	g.fillRect (0, area.y + area.height, canvas_size.width,
		    canvas_size.height - area.y - area.height);
	g.fillRect (0, area.y, area.x, area.height);
	g.fillRect (area.x + area.width, area.y,
		    canvas_size.width - area.x - area.width,
		    area.height);

	/*
	 * Fill the plot area with its background color.
	 */

	g.setColor (background);
	g.fillRect (area.x, area.y, area.width, area.height);

    } // clearCanvas ()

    
    /**
     * Draws a frame arround the plotting area, scaling ticks and labels
     * and axis labels. Check first if the paint thread should quit. In
     * that case a <CODE>QuitException</CODE> will be thrown.
     *
     *    @param  g   The graphical context to use.
     *
     *    @exception  QuitException   Thrown when the paint thread
     *				      should quit.
     *
    **********************************************************/
    private void  drawAxes (Graphics  g)  throws  QuitException
    /*********************************************************/
    {
	paint_quit.checkThrow ();    // Check quit of paint thread.

	g.setPaintMode ();
	  
	/*
	 * Draw frame around plot.
	 */

	g.setColor (getForeground ());
	g.drawRect (area.x, area.y, area.width, area.height); // Frame
	  
	FontMetrics  font	 = g.getFontMetrics ();
	int	     font_height = font.getHeight   ();
	int	     height      = font.getAscent   ();

	/*
	 * X axis ticks and labels.
	 */

	for (int i = 0;  i < x_ticks.size ();  i++)
	{
	    /*
	     * Get the X-tick from the vector. Also get the step to the
	     * next tick (or previous tick when the last is reached).
	     * The step may be used as scaler to convert the tick value
	     * to one or more tick label lines by the "DataHandler"
	     * interface.
	     */

	    Double     x_obj   = (Double) x_ticks.elementAt (i);
	    double     x       = x_obj.doubleValue      ();
	    double     x_step  = getStep	        (x_ticks, i);
	    String []  x_lines = x_handler.getTickLines (x, x_step);
	    
	    int  px = xPixel (x);
	    int  py = area.y + area.height;
	    
	    g.drawLine (px, py, px, py - tick_length);

	    // To avoid overlap with y axis labels.

	    int  space = font_height / 2;
	    int  yf    = py + height + space;

	    /*
	     * Draw the X-tick label line(s).
	     */

	    for (int j = 0;  j < x_lines.length;  j++)
	    {
		int  width = font.stringWidth (x_lines [j]);

		g.drawString (x_lines [j], px - width / 2, yf);

		yf += font_height;
	    }

	    py = area.y;
	    g.drawLine   (px, py, px, py + tick_length);
	}

	/*
	 * Y axis ticks and labels.
	 */

	for (int i = 0;  i < y_ticks.size ();  i++)
	{
	    Double     y_obj   = ((Double) y_ticks.elementAt (i));
	    double     y       = y_obj.doubleValue      ();
	    double     y_step  = getStep	        (y_ticks, i);
	    String []  y_lines = y_handler.getTickLines (y, y_step);

	    int  px = area.x;
	    int  py = yPixel (y);
	    
	    g.drawLine (px, py, px + tick_length ,py);

	    int  fy = py + height / 2 -
		      (y_lines.length - 1) * font_height;

	    for (int j = 0;  j < y_lines.length;  j++)
	    {
		int  width  = font.stringWidth (y_lines [j]);

		g.drawString (y_lines [j], px - width - 2, fy);

		fy += font_height;
	    }

	    px = area.x + area.width;
	    g.drawLine (px, py, px - tick_length, py);
	}

	if (x_axis_label != null) // x axis description
	{
	    int  s = font_height;
	    int  h = height;
	    int  w = font.stringWidth (x_axis_label);
	    int  x = area.x + area.width / 2 - w / 2;
	    int  y = area.y + area.height + s / 2 +
		     s * x_handler.getNrTickLines () + h;

	    g.drawString (x_axis_label, x, y);
	}

	if (y_axis_label != null) // y axis description
	{
	    int  s = font_height + font_height / 2 - height;
		     // = font_height;
	    int  w = font.stringWidth (y_axis_label);
	    int  x = Math.max (area.x - w - 2, OUTER_AREA_BORDER);// 0);
	    int  y = area.y - s; // / 2;

	    g.drawString (y_axis_label, x, y);
	}

    } // drawAxes ()


    /**
     * Get the step to the next tick value or to the previous tick value
     * when the last tick is reached. A zero step is returned when there
     * are less than two tick values.
     *
     *    @param  values   The vector with tick values.
     *
     *	  @param  index    The index of the tick.
     *
     *	  @return   The tick step.
     *
    ***************************************/
    private double  getStep (Vector  values,
			     int     index)
    /**************************************/
    {
	Double  value_obj;

	if (index < values.size () - 1)    // Step to next tick.
	{
	    value_obj = (Double) values.elementAt (index + 1);

	    return value_obj.doubleValue ();
	}
	else if (index > 0)    // Step to previous tick.
	{
	    value_obj = (Double) values.elementAt (index - 1);

	    return value_obj.doubleValue ();
	}
	else    // Less than two ticks: zero step.
	{
	    return 0.0;
	}

    } // getStep ()


    /**
     * Show the plot status message. This method is part of the
     * <CODE>PlotStatusHandler</CODE> interface. This implementation
     * does nothing.
     *
     *    @param  message	 The plot status message to show.
     *
     *    @param  canvas_state   The enter, be within, or exit the plot
     *				 canvas state.
     *
     *    @see  PlotStatusHandler
     *    @see  PlotCanvas#ENTER
     *    @see  PlotCanvas#WITHIN
     *    @see  PlotCanvas#EXIT
     *
    **********************************************/
    public void  showMessage (String  message,
			      int     canvas_state)
    /*********************************************/
    {
    }


    /**
     * Get the #&nbsp;lines of the tick labels. This is the default
     * implementation by this class of this method from the
     * <CODE>DataHandler</CODE> interface. It is intended for
     * &quot;normal&quot; floats.
     *
     *    @return   The #&nbsp;lines of the tick labels.
     *
     *    @see  DataHandler
     *
    ***************************/
    public int getNrTickLines ()
    /**************************/
    {
	return NR_DATA_LINES;
    }


    /**
     * Return the tick label line where the specified value is
     * converted to the requested labels. The specified step to the
     * next tick value is used as scaling parameter.
     * </P>
     *
     * <P>
     * This is the default implementation by this class of this method
     * from the <CODE>DataHandler</CODE> interface. It is intended for
     * &quot;normal&quot; floats.
     *
     *    @param  value   The tick value to convert.
     *
     *    @param  step    The step to the next tick value, intended for
     *			  scaling.
     *
     *    @return   The tick label lines.
     *
     *    @see  DataHandler
     *
    *********************************************/
    public String []  getTickLines (double  value,
				    double  step)
    /********************************************/
    {
	double  sgn_val;

	/*
	 * Return this value converted to a string when it is zero.
	 */

	if (value == 0.0)
	{
	    data_lines [0] = String.valueOf (value);

	    return data_lines;
	}

	/*
	 * Set the sign of the value to convert and set it to its
	 * absolute value.
	 */

	if (value > 0.0)
	    sgn_val = 1.0;
	else
	    sgn_val = -1.0;

	value = Math.abs (value);

	if (value < REL_STEP_BOUND * Math.abs (step))   // value << step
	{
	    /*
	     * Set the value to zero when it is absolute much smaller
	     * than its stepsize.
	     */

	    data_lines [0] = "0";
	}
	else    // |value| ~ |step|
	{
	    /*
	     * Scale the value in correspondence with the specified
	     * scale value and convert it to string.
	     */

	    double  val_scl = VALUE_SCALE;
	    double  val_pow = Math.floor (log10 (value));
	    double  val_dec = Math.pow   (10.0, val_pow);

	    value = sgn_val * Math.round (val_scl * value / val_dec) *
		    val_dec / val_scl;

	    data_lines [0] = String.valueOf (value);
	}

	return data_lines;    // Return the tick label line.

    } // getTickLines ()


    /**
     * Draws the traces inside the "plotting area".
     * Side effect: sets the clipping mask to the size of the plotting
     * area so everything outside sould be drawn before this procedure
     * is called. Check before each plot trace will be drawn if the
     * paint thread should quit. In that case a
     * <CODE>QuitException</CODE> will be thrown.
     *
     *    @param  g   The graphical context to use.
     *
     *    @exception  QuitException   Thrown when the paint thread
     *				      should quit.
     *
    **********************************************************/
    private void  plotData (Graphics  g)  throws  QuitException
    /*********************************************************/
    {
	g.setPaintMode ();

	if (use_markers)    // Markers are used.
	{
	    /*
	     * Set the clip area around the plot, but leave extra space
	     * for the markers. Note that the marker position is its
	     * centre.
	     */

	    int  marker_size = PlotMarker.getSize ();
	    int  x_clip	 = area.x	   - marker_size / 2;
	    int  y_clip	 = area.y	   - marker_size / 2;
	    int  width_clip  = area.width  + marker_size;
	    int  height_clip = area.height + marker_size;

	    /*
	     * Correct the adjusted plot area rectangle when we are out
	     * of the plot canvas.
	     */

	    if (x_clip < 0)
		x_clip = 0;

	    if (y_clip < 0)
		y_clip = 0;

	    if (width_clip > canvas_size.width)
		width_clip = canvas_size.width;

	    if (height_clip > canvas_size.height)
		height_clip = canvas_size.height;

	    /*
	     * Set the for markers adjusted plot area as the clip
	     * rectangle: it will be intersected with an existing clip
	     * rectangle.
	     */

	    g.clipRect (x_clip, y_clip, width_clip, height_clip);
	}
	else    // No markers are used: use the plot area as clip rect.
	{
	    g.clipRect (area.x, area.y, area.width, area.height);
	}

	/*
	 * Paint the plot traces. Check before a trace will be plotted
	 * if the paint thread should quit.
	 */
	  
	for (int i = 0;  i < data.size ();  i++)
	{
	    paint_quit.checkThrow ();    // quit of paint thread.

	    Vector  trace = (Vector) data.elementAt (i);

	    /*
	     * Set the trace index converted from the data index, using
	     * the "PlotTraceHandler" interface.
	     */

	    int  i_trace = trace_handler.getTraceIndex (i);

	    /*
	     * Set the plot trace color.
	     */
	    
	    g.setColor ((Color) colors.elementAt (i_trace %
						  colors.size()));

	    /*
	     * Join points?
	     */

	    boolean  join_points;
	    
	    if (use_join)
		join_points = ( (Boolean) join.elementAt (
					     i_trace % join.size ())
			      ).booleanValue ();
	    else
		join_points = false;

	    /*
	     * Use markers?
	     */

	    PlotMarker  marker;
	    
	    if (use_markers)
		marker = (PlotMarker) markers.elementAt (
					  i_trace % markers.size ());
	    else
		marker = PlotMarker.NONE_MARKER;

	    /*
	     * Use the default marker when no points are joined and no
	     * markers are used for the current plot trace.
	     */

	    if (!join_points  &&  marker.getValue () == PlotMarker.NONE)
		marker = PlotMarker.DEFAULT_MARKER;

	    /*
	     * Set offsets.
	     */

	    double  xo = ( (Double) x_offsets.elementAt (
					i_trace % x_offsets.size())
			 ).doubleValue ();
	    double  yo = ( (Double) y_offsets.elementAt (
					i_trace % y_offsets.size())
			 ).doubleValue ();

	    Point  p1 = null;
	    Point  p2; 
	  
	    if (join_points)
		for (int j = 0;  j < trace.size ();  j++) // draw trace
		{
		    /*
		     * Get the data point with the specified index.
		     */
		     
		    DataPoint  p = (DataPoint) trace.elementAt (j);

		    /*
		     * Interrupt the line when the data point is null.
		     * Go hereafter to the next point.
		     */

		    if (p == null)
		    {
			p1 = null;    // Remove previous point to join.
			continue;
		    }
		    
		    p2 = pixel (p.x + xo, p.y + yo);

		    if (p1 == null)
			p1 = p2;

		    g.drawLine (p1.x, p1.y, p2.x, p2.y);

		    p1 = p2;
		}

	    // Draw error bars (before traces?).

	    for (int j = 0;  j < trace.size ();  j++)
	    {
		DataPoint  p = (DataPoint) trace.elementAt (j);

		if (p == null)
		    continue;

		if (p.dxh != 0.0  &&  p.dxl != 0.0)
		{
		    p1 = pixel (p.x - p.dxl + xo, p.y + yo);
		    p2 = pixel (p.x + p.dxh + xo, p.y + yo);

		    g.drawLine (p1.x, p1.y, p2.x, p2.y);
		}

		if (p.dyh != 0.0  &&  p.dyl != 0.0)
		{
		    p1 = pixel (p.x + xo, p.y - p.dyl + yo);
		    p2 = pixel (p.x + xo, p.y + p.dyh + yo);
		    
		    g.drawLine (p1.x, p1.y, p2.x, p2.y);
		}
	    }

	    // Draw markers.

	    if (marker.getValue () != PlotMarker.NONE)
		for (int j = 0;  j < trace.size ();  j++)
		{
		    DataPoint  p = (DataPoint) trace.elementAt (j);

		    if (p == null)
			continue;

		    p1 = pixel (p.x + xo, p.y + yo);
		    
		    marker.draw (p1.x, p1.y, g);
		}

	} // Paint plot traces.

    } // plotData ()


    /**
     * Supply a plot trace index, given the specified data index.
     * Default the specified data index is returned.
     *
     *    @param  data_index   The specified data index.
     *
     *    @return   The index of the plot trace.
     *
     *    @see PlotTraceHandler
     *
    ******************************************/
    public int  getTraceIndex (int  data_index)
    /*****************************************/
    {
	return data_index;
    }

    
    /**
     * Convert data coordinates to screen coordinates ((0,0) is top left
     * corner).
     *
     *    @param  x   The specified <I>X</I>&nbsp;data coordinate.
     *
     *    @param  y   The specified <I>Y</I>&nbsp;data coordinate.
     *
     *    @return   The converted pixel point.
     *
    *******************************/
    private Point  pixel (double  x,
			  double  y)
    /******************************/
    {
	return new Point (xPixel (x), yPixel (y));
    }


    /**
     * Convert a <I>X</I>&nbsp;data coordinate to a pixel coordinate
     * (<I>X</I>&nbsp;=&nbsp;0 is at the left).
     *
     *    @param  x   The specified <I>X</I>&nbsp;data coordinate.
     *
     *    @return   The <I>X</I>&nbsp;pixel coordinate.
     *
    ******************************/
    private int  xPixel (double  x)
    /******************************/
    {
	double  fraction = ( xScale (x   ) - xScale (xmin) ) /
			   ( xScale (xmax) - xScale (xmin) );

	return (area.x + (int) Math.round (fraction * area.width));
    }

    
    /**
     * Convert a <I>Y</I>&nbsp;data coordinate to a pixel coordinate
     * (<I>Y</I>&nbsp;=&nbsp;0 is at the top).
     *
     *    @param  y   The specified <I>Y</I>&nbsp;data coordinate.
     *
     *    @return   The <I>Y</I>&nbsp;pixel coordinate.
     *
    ******************************/
    private int  yPixel (double  y)
    /*****************************/
    {
	double  fraction = ( yScale (y   ) - yScale (ymin) ) /
			   ( yScale (ymax) - yScale (ymin) );

	return (area.y + area.height -
		(int) Math.round (fraction * area.height));
    }
    

    /**
     * <I>X</I>&nbsp;axis scaling function, log or linear.
     *
     *    @param  x   The <I>X</I>&nbsp;value to scale.
     *
     *    @return   The scaled <I>X</I>&nbsp;value.
     *
    *********************************/
    private double  xScale (double  x)
    /********************************/
    {
	if (x_scaling == LOG)
	    if(x > 0.0)
		return log10 (x);
	    else
		return Double.NaN;
	else
	    return x;
    }


    /**
     * <I>Y</I>&nbsp;axis scaling function, log or linear.
     *
     *    @param  y   The <I>Y</I>&nbsp;value to scale.
     *
     *    @return   The scaled <I>Y</I>&nbsp;value.
     *
    *********************************/
    private double  yScale (double  y)
    /********************************/
    {
	if (y_scaling == LOG)
	    if(y > 0)
		return log10 (y);
	    else
		return Double.NaN;
	else
	    return y;
    } 

    
    /**
     * Convert an <I>X</I>&nbsp;screen coordinate to a data coordinate.
     *
     *    @param  x   The <I>X</I>&nbsp;screen coordinate to convert.
     *
     *    @return   The <I>X</I>&nbsp;data coordinate.
     *
    ******************************/
    private double  xValue (int  x)
    /*****************************/
    {
	double  fraction = (double) (x - area.x) / (double) area.width;

	return invXScale (xScale (xmin) +
			  (xScale (xmax) - xScale (xmin)) * fraction);
    }

    
    /**
     * Convert an <I>Y</I>&nbsp;screen coordinate to a data coordinate.
     *
     *    @param  y   The <I>Y</I>&nbsp;screen coordinate to convert.
     *
     *    @return   The <I>Y</I>&nbsp;data coordinate.
     *
    ******************************/
    private double  yValue (int  y)
    /*****************************/
    {
	double  fraction = (double) (area.y + area.height - y) /
			   (double) area.height;

	return invYScale (yScale (ymin) +
			  (yScale (ymax) - yScale (ymin)) * fraction);
    }

    
    /**
     * Inverse function of <I>X</I>-axis scaling function, among others
     * needed for zoom.
     *
     *    @param  x   The <I>X</I>-value to invert.
     *
     *    @return   The inverted <I>X</I>-value.
     *
    ************************************/
    private double  invXScale (double  x)
    /***********************************/
    {
	if (x_scaling == LOG)
	    return Math.pow (10.0, x);
	else
	    return x;
    } 
    

    /**
     * Inverse function of <I>Y</I>-axis scaling function, among others
     * needed for zoom.
     *
     *    @param  y   The <I>Y</I>-value to invert.
     *
     *    @return   The inverted <I>Y</I>-value.
     *
    ************************************/
    private double  invYScale (double  y)
    /***********************************/
    {
	if (y_scaling == LOG)
	    return (double) Math.pow (10.0, y);
	else
	    return y;
    } 
    

    /**
     * Round the specified value within the specified range. This method
     * is used for axis labeling.
     * </P>
     *
     * <P>
     * This method is the default implementation of the corresponding
     * method from the <CODE>DataHandler</CODE> interface. This
     * implementation is suited for &quot;normal&quot; floats.
     *
     *    @param  value   The value to round.
     *
     *    @param  range   The range in which the rounding should be
     *			  performed.
     *
     *    @return   The rounded value.
     *
     *    @see DataHandler
     *
    ***********************************/
    public double  round (double  value,
			  double  range)
    /**********************************/
    {
	if (range == 0.0)   // Return this value when the range is zero.
	    return value;

	/*
	 * Store the original value for a comparison of the value shift
	 * with the allowed range, later on in this function.
	 */

	double orig_value = value;

	/*
	 * The direction of the rounding depends from the range sign.
	 * Convert the specified range to its absolute value.
	 */

	boolean  round_up = (range < 0.0 ? false : true);
      
	range = Math.abs (range);

	/*
	 * Initialise the rounding step as the smallest decimal value
	 * induced by the range.
	 */

	double  step = Math.pow (10.0, Math.floor (log10 (range)));

	/*
	 * Scale the step up, using nice multipicators when it stays
	 * smaller than the original range.
	 */

	if (step * 5.0 <= range)
	    step *= 5.0;

	if (step * 2.0 <= range)
	    step *= 2.0;

	/*
	 * Round the value according to the rounding step and
	 * direction.
	 */

	if (round_up)
	    value = Math.ceil  (value / step) * step;
	else
	    value = Math.floor (value / step) * step;

	/*
	 * Perform the following checks:
	 *
	 *   1. Make the value zero when its absolute value is too small
	 *	compared with the rounding step.
	 *
	 *   2. Otherwise when the value is shifted more as allowed,
	 *	compared with the rounding range, set it to the original
	 *	value.
	 */

	if (Math.abs (value) < REL_STEP_BOUND * step)    // 1.
	{
	    value = 0.0;
	}
	else if ( Math.abs (value - orig_value) >	 // 2.
		  REL_MAX_RANGE_STEP * range )
	{
	    value = orig_value;
	}

	return value;    // Return the rounded value.

    } // round ()


    /**
     * Return the next tick value using the specified value and step.
     *</P>
     *
     * <P>
     * This is the default implementation of this method from the
     * <CODE>Datahandler</CODE> class. It is straightforward
     *
     * <CENTER>
     * <CODE>return value&nbsp;+&nbsp;step</CODE>
     * </CENTER>
     *
     * and suited for &quot;normal&quot; floats.
     *
     *    @param  value   The current tick value.
     *
     *    @param  step	  The step to the next tick value.
     *
     *    @return   The next tick value.
     *
     *    @see  Datahandler
     *
    *******************************************/
    public double  nextTickValue (double  value,
				  double  step)
    /******************************************/
    {
	return value + step;
    }

    
    /**
     * Get the decadic logarithm, used for rounding and log scale.
     *
     *    @param  x   A number greater than zero.
     *
     *    @return   The decadic logarithm of the specified value.
     *
    ********************************/
    private double  log10 (double  x)
    /*******************************/
    {
	return Math.log (x) / LOG_10;
    }

} // class PlotCanvas
