/* ------------------------------------------------------------------ */
/*			 S i g n a l . j a v a			      */
/*			 =====================			      */
/*								      */
/*   This file contains the source code of the main class	      */
/*   ("Signal") of the "user.signal" package.			      */
/* -------------------------------------------------------------------*/


package user.signal;

import java.awt.*;
import user.common.*;

/**
 * This class is used to display a signal light including border. It can
 * be set to the colors: wait color, alert color, go color and light off
 * color. Also the blinking of the signal can be turned on and off.
 *
 *    @author  Hans Blom
 *
**********************************************************/
public class  Signal  extends  Canvas  implements  Runnable
/*********************************************************/
{

    /*
     * Set the constant colors which can be used in the signal light to
     * show the corresponsing state.
     */

    /**
     * Set the color used at the wait signal.
     */

    public static final Color  WAIT_COLOR = Color.red;

    /**
     * Set the color used at the alert signal.
     */

    public static final Color  ALERT_COLOR = Color.yellow;

    /**
     * Set the color used at the go (save) signal.
     */

    public static final Color  GO_COLOR = Color.green;

    /*
     * Set the color which serves as a border around the light.
     */

    private static final Color  SIGNAL_COLOR = Color.black;

    /*
     * Set the color used when the signal is turned off.
     */

    private static final Color  OFF_COLOR = Color.gray;

    /*
     * Set the factor by which the default blinking time will be devided
     * to obtain the blinking repaint time.
     */

    private static final int  BLINK_REPAINT_TIME_FACTOR = 10;

    /*
     * Set the class instance variables.
     */

    private Image	   buffer_image;  // Double buffer image.
    private Graphics	   g_image;	  // Graphic context buffer im.
    private Color	   light_color;   // Color of the signal light.
    private Color	   paint_color;	  // Signal color to be painted.
    private BlinkChecker   blink_checker; // Signal blink ckecker.
    private Thread	   blink_thread;  // Signal blink thread.
    private QuitSync	   blink_quit;	  // Quit from blink thread?
    private BooleanObject  do_blink_obj;  // Signal does blonk object?
    private int		   blink_time;	  // Blink time [ms].
    private int		   blink_update_t;// Max. updt. t. at bl. paint.
    private int		   width;	  // Width of the signal canvas.
    private int		   height;	  // Height of the sign. canvas.
    private int		   signal_coord;  // Signal coordinate (X == Y).
    private int		   signal_size;   // Size of the signal.
    private int		   signal_border; // Border around the signal.
    private int		   light_coord;   // X & Y coord. of the light.
    private int		   light_size;    // Size of the signal light.
    private int		   light_border;  // Border around the light.
    private BooleanObject  painted_obj;	  // The signal is yet painted?


    /**
     * Constructor.
     *
     *    @param  light_color     The initial color of the signal light.
     *
     *    @param  blink_checker   The interface which should check if an
     *				  off signal should be painted.
     *
     *	  @param  blink_time	  The blinking time [ms].
     *
     *    @param  signal_border   The border around the signal.
     *
     *    @param  light_border    The border around the signal light.
     *
    *******************************************/
    public  Signal (Color	  light_color,
		    BlinkChecker  blink_checker,
		    int		  blink_time,
		    int		  signal_border,
		    int		  light_border)
    /******************************************/
    {
	/*
	 * Initialise the class instance variables.
	 */

	this.buffer_image   = null;
	this.g_image	    = null;
	this.light_color    = light_color;
	this.paint_color    = light_color;
	this.blink_thread   = new Thread   (this);
	this.blink_quit     = new QuitSync ();
	this.blink_checker  = blink_checker;
	this.do_blink_obj   = new BooleanObject (false);
	this.blink_time	    = blink_time;
	this.blink_update_t = blink_time / BLINK_REPAINT_TIME_FACTOR;
	this.width	    = 0;
	this.height	    = 0;
	this.signal_coord   = 0;
	this.signal_size    = 0;
	this.signal_border  = signal_border;
	this.light_coord    = 0;
	this.light_size	    = 0;
	this.light_border   = light_border;
	this.painted_obj    = new BooleanObject (false);
    
    } // Signal ()


    public void  setBlinking (boolean  do_blink)
    {
	synchronized (do_blink_obj)
	{
	    do_blink_obj.value = do_blink;
	}
    }


    public void  start ()
    {
	if (blink_thread.isAlive ())
	    blink_thread.resume ();
	else
	    blink_thread.start ();
    }


    public void  stop ()
    {
	if (blink_thread.isAlive ())
	    blink_thread.suspend ();
    }


    public void  destroy ()
    {
	if (blink_thread.isAlive ())
	{
	    blink_quit.set	();
	    blink_thread.resume ();
	}
    }


    /**
     * This function implements the <CODE>run()</CODE> method used in
     * the load blink thread. It shows a blinking alert signal when the
     * loading of the net data is sufficient fast. Jump from this method
     * when a <CODE>QuitException</CODE> has been thrown by the blinking
     * thread. It will be thrown when the thread should stop.
     *
    ******************/
    public void  run ()
    /*****************/
    {
	/*
	 * Catch the Quit Exception which will be thrown when the blink
	 * thread should stop.
	 */

	try
	{
	    /*
	     * Loop forever as long as the Applet exists (i.e. no Quit
	     * Exceptions are thrown). Externally this thread may be
	     * "suspend()"-ed ("resumed()"-ed) by the "stop()"
	     * ("start()") methods.
	     */

	    for (;;)
	    {
		/*
		 * Sleep before switching the signal light off, when
		 * required.
		 */

		try
		{
		    Thread.sleep (blink_time);
		}
		catch (InterruptedException  e_intr_off)
		{
		}

		/*
		 * Set some help variables reflecting the current state
		 * of
		 * the synchronised variables listed below.
		 */

		boolean  painted;   // The signal is already painted.

		synchronized (painted_obj)
		{
		    painted = painted_obj.value;
		}

		boolean  do_blink;    // Blinking is enabled.

		synchronized (do_blink_obj)
		{
		    do_blink = do_blink_obj.value;
		}

		/*
		 * Jump from the blink loop when a blink thread quit has
		 * been selected just before the signal may be painted
		 * off.
		 */

		blink_quit.checkThrow ();

		/*
		 * Switch the light off when:
		 *
		 *    1. The signal has already been painted.
		 *
		 *    2. Blinking is enabled.
		 *
		 *    3. The blick checker conditions are fullfilled.
		 */

		if ( painted			   /* 1. */  &&
		     do_blink			   /* 2. */  &&
		     blink_checker.isFulfilled ()  /* 3. */ )
		{
		    paintLight (OFF_COLOR);    // Switch light off.
		}

		/*
		 * Sleep before switching the signal light on with the
		 * current light color, when required.
		 */

		try
		{
		    Thread.sleep (blink_time);
		}
		catch (InterruptedException  e_intr_on)
		{
		}

		/*
		 * Jump from the blink loop when a blink thread quit has
		 * been selected just before the signal may be painted
		 * off.
		 */

		blink_quit.checkThrow ();

		/*
		 * Only switch the light on when the signal has already
		 * been painted.
		 */

		if (painted)    // The signal has been painted.
		{
		    /*
		     * Obtain the current painted and light color in a
		     * synchronised way with the Applet.
		     */

		    Color  curr_light_color;
		    Color  curr_paint_color;

		    synchronized (painted_obj)
		    {
			curr_light_color = light_color;
			curr_paint_color = paint_color;
		    }

		    /*
		     * Only paint the light agaon on when the current
		     * light color is not equal to the paint color. This
		     * is always the case when the light has been
		     * switched of during blinking.
		     */

		    if (!curr_light_color.equals (curr_paint_color))
			paintLight (curr_light_color);

		} // The signal has been painted.

	    } // Loop forever.

	}
	catch (QuitException  quit_except)    // Stop the thread when a
	{				      // quit has been selected.
	}

    } // run ()


    /**
     * Set the color of the signal light and repaint it. The repaint is
     * only done when the color is changed and the signal was already
     * painted. The setting of the light color is synchronised with the
     * blink thread.
     *
     *    @param  light_color   The new color of the the light.
     *
    **********************************************/
    public void  setLightColor (Color  light_color)
    /*********************************************/
    {
	synchronized (painted_obj)
	{
	    this.light_color = light_color;
	}

	paintLight (light_color);
    }


    /**
     * Set the color which should be used to paint the signal in a
     * synchronised way and repaint it. The painting is only done when
     * the color is changed and the signal was already painted.
     * painted.
     *
     *    @param  paint_color   The new color to paint.
     *
    ********************************************/
    private void  paintLight (Color  paint_color)
    /*******************************************/
    {
	/*
	 * Check the paint color synchronised with the blink thread.
	 */

	boolean  is_same_color;    // Unchanged color flag.
	boolean  painted;	   // Signal has yet been painted flag.

	synchronized (painted_obj)   // Synchronised with blink thread.
	{
	    /*
	     * Store the unchamged state of the color and the state if
	     * the color has been changed in help varaibles to be used
	     * outside of this synchronised region.
	     */

	    is_same_color = this.paint_color.equals (paint_color);
	    painted	  = painted_obj.value;

	    /*
	     * Set the new paint color when it has been changed.
	     */

	    if (!is_same_color)
		this.paint_color = paint_color;

	} // Synchronised with blink thread.

	/*
	 * Repaint the signal when it has already been painted and when
	 * the paint color has been changed.
	 */

	if (painted && !is_same_color)    // Repaint the signal.
	{
	    /*
	     * Use the maximum update time when blinking is enabled; no
	     * update time is used otherwise. Retrieve the enable blink
	     * state in a synchronised way between the various threads.
	     */

	    boolean  do_blink;   // Help variable for blink enable flag.

	    synchronized (do_blink_obj)
	    {
		do_blink = do_blink_obj.value;
	    }

	    if (do_blink)    // Blinking enabled: use max. update time.
	    {
		repaint (blink_update_t);
	    }
	    else    // No blinking, so no maximum update time.
	    {
		repaint ();
	    }

	} // Repaint the signal.

    } // paintLight ()


    /**
     * Update the graphics. This can be done directly with a
     * complete <TT>paint()</TT> of the signal without a clear, because
     * double buffering is used.
     *
     *    @param  g   The graphical context.
     *
    ********************************/
    public void  update (Graphics  g)
    /*******************************/
    {
	paint (g);
    }


    /**
     * Paint the signal. Double buffering is used to prevend flicker.
     * The color of the signal to paint is retrieved in a synchronised
     * way between the various threads.
     *
     *    @param  g   The graphical context.
     *
    *******************************/
    public void  paint (Graphics  g)
    /******************************/
    {
	/*
	 * Retrieve the current paint color in a synchronised way
	 * between the various threads.
	 */

	Color  curr_paint_color;

	synchronized (painted_obj)
	{
	    curr_paint_color = paint_color;
	}

	/*
	 * Do the painting in a synchronised way between the various
	 * threads, using the specal synchronised paint method.
	 */

	paintSynchronized (g, curr_paint_color);

	/*
	 * Store that the signal has been painted in a synchronised way
	 * between the various threads.
	 */

	synchronized (painted_obj)
	{
	    painted_obj.value = true;    // Signal has been painted.
	}

    } // paint ()


    /**
     * Paint the signal. Double buffering is used to prevend flicker.
     * The color of the signal to paint is retrieved in a synchronised
     * way by this method between the various threads.
     *
     *    @param  g		     The graphical context.
     *
     *    @param  curr_paint_color   The current (non-synchronised)
     *				     paint color of the signal light.
     *
    **********************************************************/
    private synchronized void  paintSynchronized (
				   Graphics  g,
				   Color      curr_paint_color)
    /*********************************************************/
    {
	/*
	 * Check if the sizes are changed. If so change the appropriate
	 * variables.
	 */

	checkResized ();

	/*
	 * Draw the signal in the image buffer.
	 */

	g_image.setColor (getBackground ());	  // Clear image with
	g_image.fillRect (0, 0, width, height);   // current background.
	g_image.setColor (SIGNAL_COLOR);
	g_image.fillOval (signal_coord, signal_coord,    // Draw border.
			  signal_size,  signal_size);
	g_image.setColor (curr_paint_color);
	g_image.fillOval (light_coord, light_coord,    // Draw light.
			  light_size,  light_size);
	
	/*
	 * Draw the image in the display.
	 */

	g.drawImage (buffer_image, 0, 0, this);

    } // paintSynchronized ()


    /**
     * Check if this component is resized. If so (re)allocate the
     * buffer image and adjust the signal coordinates and sizes.
     *
    ****************************/
    private void  checkResized ()
    /***************************/
    {
	/*
	 * Inquire the current size.
	 */

	Dimension  curr_dim = size ();

	/*
	 * Return when nothing is changed.
	 */

	if (curr_dim.width == width  &&  curr_dim.height == height)
	    return;

	/*
	 * Set the new sizes and (re)allocate the buffer image. Also
	 * obtain the graphical context from the new buffer image.
	 */

	width	     = curr_dim.width;
	height	     = curr_dim.height;
	buffer_image = createImage		(width, height);
	g_image	     = buffer_image.getGraphics ();

	/*
	 * Set the signal size to the smallest of both dimensions.
	 */

	if (width < height)
	    signal_size = width;
	else
	    signal_size = height;

	/*
	 * Adjust the sizes and origin coordinate for the border of the
	 * signal. Set the border to zero when the signal size is too
	 * small.
	 */

	if (signal_size > 2 * signal_border) // Adjust for signal bord.
	{
	    signal_coord  = signal_border;
	    signal_size  -= (2 * signal_border);
	}
	else    // Signal size too small: set signal border to zero.
	{
	    signal_coord = 0;
	}

	/*
	 * Set the light size by distracting the light border from the
	 * signal size. Also set this border to zero when the signal
	 * size is too small.
	 */

	if (signal_size > 2 * light_border)  // Adjust for light border.
	{
	    light_coord = signal_coord + light_border;
	    light_size  = signal_size  - 2 * light_border;
	}
	else    // Signal size too small: set light border to zero.
	{
	    light_coord = signal_coord;
	    light_size  = signal_size;
	}
    
    } // checkResized ()
    
}  // class Signal
