#! /usr/bin/perl -w

use strict;
use Socket qw(:DEFAULT :crlf);
use POSIX;

my $FALSE = 0;
my $TRUE  = 1;

my $UNIX    = 1;
my $WINDOWS = 2;

my $SYSTEM_TYPE;
my $DIR_SEPAR;

if ($ENV{'PATH'} =~ /;/)
{
    $SYSTEM_TYPE = $WINDOWS;
    $DIR_SEPAR	 = "\\";
}
else
{
    $SYSTEM_TYPE = $UNIX;
    $DIR_SEPAR	 = "/";
}

my $PROGRAM  = basename ($0);
my $PROG_DIR = dirname	($0);

my $HOME_DIR;

if ($SYSTEM_TYPE == $UNIX)
{
    $HOME_DIR = $ENV{'HOME'};
}
else    # $WINDOWS
{
    $HOME_DIR = "$PROG_DIR\\..";
}

sub main
{
    my $argv = shift;

    my $param = Param->new ();

    $param->parseArguments ($argv);

    my $net_test_com_server = NetTestComServer->new ($param);

    $net_test_com_server->run ();
}

sub basename
{
    my $path = shift;

    $path =~ s/$DIR_SEPAR*$//;
    
    my @path_parts;
    
    if ($SYSTEM_TYPE == $UNIX)
    {
	@path_parts = split (/\//, $path);
    }
    else    # WINDOWS
    {
	@path_parts = split (/\\/, $path);
    }

    return $path_parts [$#path_parts];
}

sub dirname
{
    my $path = shift;

    $path =~ s/$DIR_SEPAR*$//;

    if ($path =~ s/$DIR_SEPAR[^$DIR_SEPAR]*$//)
    {
	return $path;
    }
    else
    {
	return '.';
    }
}

package Param;

use vars qw ( $DFT_PORT

	      $DEBUG_KEY
	      $LOG_FILE_KEY
	      $PID_FILE_KEY
	      $HOST_ALLOW_HASH_KEY	
	      $ACCEPT_ALL_KEY
	      $PORT_KEY );

$DFT_PORT = 37373;

$DEBUG_KEY	     = 'debug';
$LOG_FILE_KEY	     = 'log_file';
$PID_FILE_KEY	     = 'pid_file';
$HOST_ALLOW_HASH_KEY = 'host_allow_hash';
$ACCEPT_ALL_KEY      = 'accept_all';
$PORT_KEY	     = 'port';

sub new
{
    my $class = shift;

    my $self = {};

    $self->{$DEBUG_KEY		} = $FALSE;
    $self->{$LOG_FILE_KEY	} = '-';
    $self->{$PID_FILE_KEY	} = undef;
    $self->{$HOST_ALLOW_HASH_KEY} = {};
    $self->{$ACCEPT_ALL_KEY	} = $TRUE;
    $self->{$PORT_KEY		} = $DFT_PORT;

    return bless ($self, $class);
}

sub parseArguments
{
    my $self = shift;
    my $argv = shift;

    while ( defined (my $arg = shift (@$argv)) )
    {
	if ($arg eq '-debug')
	{
	    $self->{$DEBUG_KEY} = $TRUE;
	}
	elsif ($arg eq '-config_file')
	{
	    $self->parseConfigFile (shift (@$argv));
	}
	elsif ($arg eq '-log_file')
	{
	    $self->{$LOG_FILE_KEY} = getParseFullPath (shift (@$argv));
	}
	elsif ($arg eq '-pid_file')
	{
	    $self->{$PID_FILE_KEY} = getParseFullPath (shift (@$argv));
	}
	elsif ($arg eq '-allow_host')
	{
	    my $host = shift (@$argv);

	    $self->{$HOST_ALLOW_HASH_KEY }->{$host} = $host;
	    $self->{$ACCEPT_ALL_KEY	 }	    = $FALSE;
	}
	elsif ($arg eq '-port')
	{
	    $self->{$PORT_KEY} = shift (@$argv);
	}
	else
	{
	    printUsage ( );
	    exit       (1);
	}
    }
}

sub parseConfigFile
{
    my $self	    = shift;
    my $config_file = shift;

    $config_file = getParseFileName ($config_file);

    local *config_read;

    open (config_read, "<$config_file")  ||
	die "$PROGRAM [$$]: $!: Open config file for reading: " .
	    "$config_file\n";

    my $config_expr = undef;
    my $line_nr     = 0;

    while (<config_read>)
    {
	$line_nr++;
	chomp;

	s/#.*$//;
	s/^\s+//;
	s/\s+$//;

	if (/=/)
	{
	    if (defined ($config_expr))
	    {
		$self->parseConfigExpr ($config_expr, $line_nr - 1);
	    }

	    $config_expr = $_;
	}
	elsif (!/^\s*$/)
	{
	    if (defined ($config_expr))
	    {
		$config_expr .= " $_";
	    }
	    else
	    {
		$config_expr = $_;
	    }
	}
    }

    if (defined ($config_expr))
    {
	$self->parseConfigExpr ($config_expr, $line_nr);
    }

    close (config_read)  ||
	die "$PROGRAM [$$]: $!: Close config file after reading: " .
	    "$config_file\n";
}

sub parseConfigExpr
{
    my $self	    = shift;
    my $config_expr = shift;
    my $line_nr     = shift;

    my ( $name, $value ) = split (/\s*=\s*/, $config_expr);

    if (!defined ($value))
    {
	die "$PROGRAM [$$]: Bad configuation expression at line " .
	    "#$line_nr: $config_expr\n";
    }

    $self->parseArguments ( [ "-$name", split (/\s+/, $value) ] );
}

sub getParseFullPath
{
    my $path = shift;

    $path = getParseFileName ($path);

    if ($SYSTEM_TYPE == $WINDOWS)
    {
	return $path;
    }

    if ($path =~ /^\//)
    {
	return $path;
    }
    else
    {
	chomp (my $curr_dir_path = `pwd`);

	return "$curr_dir_path/$path\n";
    }
}

sub getParseFileName
{
    my $file_name = shift;

    $file_name =~ s/\$HOME/$HOME_DIR/;
    $file_name =~ s/\$PROGRAM/$PROGRAM/;

    return $file_name;
}

sub printUsage
{
    print STDERR
	  "Usage: $PROGRAM [<option> [...]]\n",
	  "\n",
	  "Possible options:\n",
	  "\n",
	  "    {-help,-?}\n",
	  "          Print this message and exit.\n",
	  "\n",
	  "    -debug\n",
	  "          Print extra log debug information\n",
	  "\n",
	  "    -config_file {<config_file>,-}\n",
	  "          Specify a configuration file to supply the ",
	  "program\n",
	  "          parameters. They may be given in the format:\n",
	  "\n",
	  "              <name> = <value>\n",
	  "\n",
	  "          where <name> is one of the program options ",
	  "without\n",
	  "          leading '-'. The <value> fields may be spread ",
	  "across multiple\n",
	  "          lines. No line escape character is used. Lines ",
	  "with spaces\n",
	  "          and comment text after '#' are ignored. When '-' ",
	  "is\n",
	  "          specified, Standard Input will be used. The ",
	  "following\n",
	  "          variables are substituted by:\n",
	  "\n";

    if ($SYSTEM_TYPE == $UNIX)
    {
	print STDERR
	      "              \$HOME      The home directory of the ",
	      "user.\n";
    }
    else    # $WINDOWS
    {
	print STDERR
	      "              \$HOME      The sup directory of this ",
	      "script.\n";
    }

    print STDERR
	  "\n",
	  "              \$PROGRAM   The base program name.\n",
	  "\n",
	  "    -log_file {<log_file>,-}\n",
	  "          Specify a log file. The same variables as in ",
	  "the\n",
	  "          \"-config_file\" option may be used.";
	  
    if ($SYSTEM_TYPE == $UNIX)
    {
	print STDERR
	      " When no log file, or '-'\n",
	      "          is specified, this server will run in the ",
	      "foreground.\n",
	      "\n",
	      "    -pid_file <pid_file>\n",
	      "          Specify a file to which the process ID of ",
	      "the daemon server\n",
	      "          (when used) will be written. The same ",
	      "variables as in the\n",
	      "          \"-config_file\" option may be used. Default ",
	      "no such file will\n",
	      "          be used.\n";
    }
    else    # $WINDOWS.
    {
	print STDERR
	      "\n";
    }

    print STDERR
	  "\n",
	  "    -allow_host <host>\n",
	  "          Allow only connections from the specified host. ",
	  "This option\n",
	  "          may be multiple specified. Default connections ",
	  "from all hosts\n",
	  "          are allowed.\n",
	  "\n",
	  "    -port <port>\n",
	  "          Specify the port where this server listens. The ",
	  "default\n",
	  "          is: $DFT_PORT.\n";
}

sub getDebug
{
    my $self = shift;

    return $self->{$DEBUG_KEY};
}

sub getPort
{
    my $self = shift;

    return $self->{$PORT_KEY};
}

sub acceptAllHosts
{
    my $self = shift;

    return $self->{$ACCEPT_ALL_KEY};
}

sub acceptHost
{
    my $self = shift;
    my $host = shift;

    my $accept_all	= $self->{$ACCEPT_ALL_KEY     };
    my $host_allow_hash = $self->{$HOST_ALLOW_HASH_KEY};

    return ( $accept_all  ||  defined ($host_allow_hash->{$host}) );
}

sub useLogFile
{
    my $self = shift;

    my $log_file = $self->{$LOG_FILE_KEY};

    return (defined ($log_file)  &&  $log_file ne '-');
}

sub getLogFile
{
    my $self = shift;

    return $self->{$LOG_FILE_KEY};
}

sub getPIDFile
{
    my $self = shift;

    return $self->{$PID_FILE_KEY};
}

package NetTestComServer;

use vars qw ( $MAX_NR_LOG_FILE_LINES
	      $CHECK_LOG_INTV

	      $WAITED_PID

	      $PARAM_KEY );

$MAX_NR_LOG_FILE_LINES = 5000;
$CHECK_LOG_INTV        =   20;

$WAITED_PID = 0;

$PARAM_KEY = 'param';

sub new
{
    my $class = shift;
    my $param = shift;

    my $self = {};

    $self->{$PARAM_KEY} = $param;

    return bless ($self, $class);
}

sub run
{
    my $self = shift;

    my $param = $self->{$PARAM_KEY};

    if ($param->useLogFile ())
    {
	if ($SYSTEM_TYPE == $UNIX)
	{
	    $self->makeDaemon ();
	}

	$self->openLog ();
    }

    my $listen_socket = $self->openListenSocket ();

    $self->handleIncoming ($listen_socket);

    close ($listen_socket)  ||
	die "$PROGRAM [$$]: $!: Closing listen socket\n";

    wait ();
}

sub makeDaemon
{
    my $self = shift;

    chdir ('/')  ||
	die "$PROGRAM [$$]: $!: Change to root directory: /\n";

    open (STDIN, "</dev/null")  ||
	die "$PROGRAM [$$]: $!: Redirect stdin from null device: " .
	    "/dev/null\n";

    my $daemon_pid = fork ();

    if (!defined ($daemon_pid))
    {
	die "$PROGRAM [$$]: $!: Can't fork to make this script a " .
	    "daemon\n";
    }
    elsif ($daemon_pid == 0)
    {
	main::setsid ()  ||
	    die "$PROGRAM [$$]: $!: Can't start a new session\n";
    }
    else
    {
	$self->checkWritePIDFile ($daemon_pid);

	exit (0);
    }
}

sub checkWritePIDFile
{
    my $self	   = shift;
    my $daemon_pid = shift;

    my $pid_file = $self->{$PARAM_KEY}->getPIDFile ();

    if (!defined ($pid_file))
    {
	return;
    }

    local *pid_write;

    open (pid_write, ">$pid_file")  ||
	die "$PROGRAM [$$]: $!: Open file to write daemon PID: " .
	    "$pid_file\n";

    print pid_write "$daemon_pid\n";

    close (pid_write)  ||
	die "$PROGRAM [$$]: $!: Close daemon PID file after writing: " .
	    "$pid_file\n";
}

sub openListenSocket
{
    my $self = shift;

    my $port = $self->{$PARAM_KEY}->getPort ();

    my $pack_addr = main::sockaddr_in ($port, main::INADDR_ANY);
    my $proto	  = getprotobyname    ('tcp');

    local *listen_socket;

    socket (listen_socket, main::PF_INET, main::SOCK_STREAM, $proto)  ||
	die "$PROGRAM [$$]: $!: Open listen socket\n";

    setsockopt (listen_socket,
		main::SOL_SOCKET,
		main::SO_REUSEADDR,
		pack ("l", 1))  ||
	die "$PROGRAM [$$]: $!: Set listen socket options\n";

    bind (listen_socket, $pack_addr)  ||
	die "$PROGRAM [$$]: $!: Bind listen socket\n";

    listen (listen_socket, main::SOMAXCONN)  ||
	die "$PROGRAM [$$]: $!: Listen to listen socket\n";

    my $start_time = localtime ();

    print "[$$]  $start_time:\n",
	  "[$$]      Listening on port $port\n";

    return *listen_socket{IO};
}

sub handleIncoming
{
    my $self	      = shift;
    my $listen_socket = shift;

    my $param = $self->{$PARAM_KEY};

    my $use_log_file = $param->useLogFile ();

    if ($SYSTEM_TYPE == $UNIX)
    {
	$SIG{CHLD} = \&waitChilds;
    }

    my	  $pack_addr;
    local *client_socket;
    my	  $check_log_count = $CHECK_LOG_INTV - 1;

    for ( $WAITED_PID = 0;
	  ($pack_addr = accept (client_socket, $listen_socket))
	  ||
	  $WAITED_PID != 0;
	  $WAITED_PID = 0, close (client_socket) )
    {
	if ($WAITED_PID != 0  &&  !$pack_addr)
	{
	    next;
	}

	if ($use_log_file)
	{
	    $self->closeLog ();
	    $self->openLog  ();
	}

	if ($SYSTEM_TYPE == $UNIX)
	{
	    my $net_test_pid = fork ();

	    if (!defined ($net_test_pid))
	    {
		die "$PROGRAM [$$]: $!: Fork to handle the net test\n";
	    }
	    elsif ($net_test_pid == 0)
	    {
		my $net_test_com =
		    NetTestCommand->new ($param, *client_socket{IO});

		$net_test_com->run ();

		exit (0);
	    }
	}
	else    # $WINDOWS
	{
	    my $net_test_com = NetTestCommand->new ($param,
						    *client_socket{IO});

	    $net_test_com->run ();
	}

	if ($use_log_file  &&  ++$check_log_count >= $CHECK_LOG_INTV)
	{
	    if ($SYSTEM_TYPE == $UNIX)
	    {
		my $check_log_pid = fork ();

		if (!defined ($check_log_pid))
		{
		    die "$PROGRAM [$$]: $!: Fork to check the log " .
			"file size\n";
		}
		elsif ($check_log_pid == 0)
		{
		    $self->checkLogFileSize ();

		    exit (0);
		}
	    }
	    else    # $WINDOWS
	    {
		$self->checkLogFileSize ();
	    }

	    $check_log_count = 0;
	}
    }
}

sub checkLogFileSize
{
    my $self = shift;

    my $param = $self->{$PARAM_KEY};

    if (!$param->useLogFile ())
    {
	return;
    }

    my $log_file = $param->getLogFile ();

    local *log_read;

    open (log_read, "<$log_file")  ||
	die "$PROGRAM [$$]: $!: Open log to check file for reading: " .
	    "$log_file\n";

    my $nr_log_file_lines;

    while (<log_read>)
    {
	$nr_log_file_lines++;
    }

    close (log_read)  ||
	die "$PROGRAM [$$]: $!: Close log file to check after " .
	    "reading: $log_file\n";

    if ($nr_log_file_lines > $MAX_NR_LOG_FILE_LINES)
    {
	my $old_log_file = "$log_file.old";

	rename ($log_file, $old_log_file)  ||
	    die "$PROGRAM [$$]: $!: Rename $log_file to " .
		"$old_log_file\n";
    }
}

sub openLog
{
    my $self = shift;

    my $log_file = $self->{$PARAM_KEY}->getLogFile ();

    open (STDOUT, ">>$log_file")  ||
	die "$PROGRAM: $!: Redirect append stdout to log file: " .
	    "$log_file\n";

    open (STDERR, ">&STDOUT")  ||
	die "$PROGRAM: $!: Redirect stderr to stdout\n";

    my $old_descr = select (STDOUT);  $| = 1;

    select (STDERR    );  $| = 1;
    select ($old_descr);
}

sub closeLog
{
    my $self = shift;

    close (STDOUT)  ||  die "$PROGRAM: $!: Close log stdout\n";
    close (STDERR)  ||  die "$PROGRAM: $!: Close log stderr\n";
}

sub waitChilds
{
    $WAITED_PID = wait ();

    $SIG{CHLD} = \&waitChilds;   # Required for SysV.

    if ($WAITED_PID < 0)
    {
	$WAITED_PID = 0;
    }
}

package NetTestCommand;

use vars qw ( $TIMEOUT_TIME
	      $BASE_COM_ALLOW_HASH
	      $TMP_ERROR_BASE

	      $ARGV_START
	      $ARGV_END
	      $STDOUT_START
	      $STDOUT_END
	      $STDERR_START
	      $STDERR_END
	      $EXIT_VAL_START
	      $EXIT_VAL_END

	      $PARAM_KEY
	      $CLIENT_SOCKET_KEY
	      $PEER_HOSTNAME_KEY );

$TIMEOUT_TIME = 7200; # s.

$BASE_COM_ALLOW_HASH = { 'start_netserver'   => 'start_netserver',
			 'start_udpmon_resp' => 'start_udpmon_resp',
			 'host_net_test'     => 'host_net_test' };

$TMP_ERROR_BASE = getTmpErrorBase ();

$ARGV_START	= ':Command-Argv-Start:';
$ARGV_END	= ':Command-Argv-End:';
$STDOUT_START	= ':Command-STDOUT-Start:';
$STDOUT_END	= ':Command-STDOUT-End:';
$STDERR_START	= ':Command-STDERR-Start:';
$STDERR_END	= ':Command-STDERR-End:';
$EXIT_VAL_START = ':Command-ExitVal-Start:';
$EXIT_VAL_END   = ':Command-ExitVal-End:';

$PARAM_KEY	   = 'param';
$CLIENT_SOCKET_KEY = 'client_socket';
$PEER_HOSTNAME_KEY = 'peer_hostname';

sub new
{
    my $class	      = shift;
    my $param	      = shift;
    my $client_socket = shift;

    my $self = {};

    $self->{$PARAM_KEY	      } = $param;
    $self->{$CLIENT_SOCKET_KEY} = $client_socket;

    return bless ($self, $class);
}

sub run
{
    my $self = shift;

    if ($SYSTEM_TYPE == $UNIX)
    {
	eval
	{
	    local $SIG{ALRM} = sub { die "alarm\n" };

	    alarm		     ($TIMEOUT_TIME);
	    $self->runCommandSession ();
	    alarm		     (0);
	};

	if ($@)
	{
	    if ($@ ne "alarm\n")
	    {
		die;
	    }

	    print "[$$]  Timeout in running command\n";
	}
    }
    else    # $WINDOWS
    {
	$self->runCommandSession ();
    }

    $self->closeClient ();
}

sub runCommandSession
{
    my $self = shift;

    $self->unbufferSocket ();

    if ($self->checkPeerName ())
    {
	my $argv = $self->readArgv ();

	if ($self->checkArgv ($argv))
	{
	    $self->runWrite ($argv);
	}
    }
}

sub unbufferSocket
{
    my $self = shift;

    my $client_socket = $self->{$CLIENT_SOCKET_KEY};

    my $old_descr = select ($client_socket);  $| = 1;

    select ($old_descr);
}

sub checkPeerName
{
    my $self = shift;

    my $param	      = $self->{$PARAM_KEY	  };
    my $client_socket = $self->{$CLIENT_SOCKET_KEY};

    my $peer_socket_addr = getpeername ($client_socket);

    my ($port, $peer_inet_addr) = main::sockaddr_in ($peer_socket_addr);

    my $peer_hostname = gethostbyaddr ($peer_inet_addr, main::AF_INET);

    my $peer_ip_addr_str = main::inet_ntoa ($peer_inet_addr);

    my $peer_connect_str;
    my $peer_host_str;

    if (defined ($peer_hostname))
    {
	$peer_connect_str = "$peer_hostname ($peer_ip_addr_str)";
	$peer_host_str	  = "$peer_hostname";
    }
    else
    {
	$peer_hostname	  = $peer_ip_addr_str;
	$peer_connect_str = $peer_ip_addr_str;
	$peer_host_str	  = $peer_ip_addr_str;
    }

    $self->{$PEER_HOSTNAME_KEY} = $peer_hostname;

    print "[$$]  Connecting $peer_connect_str on port $port\n";

    if ( !$param->acceptHost ($peer_hostname   )  &&
	 !$param->acceptHost ($peer_ip_addr_str) )
    {
	print "[$$]  Connection from peer $peer_host_str refused\n";

	return $FALSE;
    }

    return $TRUE;
}

sub readArgv
{
    my $self = shift;

    my $param	      = $self->{$PARAM_KEY	  };
    my $client_socket = $self->{$CLIENT_SOCKET_KEY};

    my $debug = $param->getDebug ();

    my $argv	  = [];
    my $read_argv = $FALSE;

    while (<$client_socket>)
    {
	s/^\s+//;
	s/\s+$//;

	if ($debug)
	{
	    print "[$$]  << $_\n";
	}

	if ($_ eq $ARGV_START)
	{
	    $read_argv = $TRUE;
	}
	elsif ($_ eq $ARGV_END)
	{
	    return $argv;
	}
	elsif ($read_argv  &&  !/^\s*$/)
	{
	    push (@$argv, $_);
	}
    }
}

sub checkArgv
{
    my $self = shift;
    my $argv = shift;

    if ( !defined ($argv)  ||
	 $#$argv < $[	   ||
	 !defined ($BASE_COM_ALLOW_HASH->{$argv->[$[]}) )
    {
	print "[$$]  Refusing invalid net test command: @$argv\n";

	return $FALSE;
    }

    foreach my $arg (@$argv)
    {
	if ( !($arg =~ /^[A-Za-z0-9_\-\/,.]+$/) )
	{
	    print "[$$]  Refusing net test command for invalid " .
		  "arguments: @$argv\n";

	    return $FALSE;
	}
    }

    return $TRUE;
}

sub runWrite
{
    my $self = shift;
    my $argv = shift;

    my $client_socket = $self->{$CLIENT_SOCKET_KEY};

    $argv->[0] = $PROG_DIR . $DIR_SEPAR . $argv->[0];

    print "[$$]  Running: @$argv\n";

    my $pid;
    my $exit_val;

    if ($SYSTEM_TYPE == $UNIX)
    {
	$self->runWriteUnix ($argv);
    }
    else    # $WINDOWS
    {
	$self->runWriteWindows ($argv);
    }
}

sub runWriteUnix
{
    my $self = shift;
    my $argv = shift;

    my $param	      = $self->{$PARAM_KEY	  };
    my $client_socket = $self->{$CLIENT_SOCKET_KEY};

    my $debug = $param->getDebug ();

    local *output_read;
    local *output_write;

    pipe (output_read, output_write)  ||
	die "$PROGRAM [$$]: $!: Create pipe ro run command: @$argv\n";

    $SIG{CHLD} = &dummyFunction;

    my $pid = fork ();

    if (!defined ($pid))
    {
	die "$PROGRAM [$$]: $!: Fork to run command: @$argv\n";
    }
    elsif ($pid == 0)
    {
	$SIG{INT} = \&unlinkTmpErrorFileAndExit;

	my $tmp_error_file = getTmpErrorFile ($$);

	close (output_read)  ||
	    die "$PROGRAM [$$]: $!:Close output read pipe in child " .
		"to run: @$argv\n";

	open (STDIN, "</dev/null")  ||
	    die "$PROGRAM [$$]: $!: Redirect stdin from /dev/null in " .
		"child to run: @$argv\n";

	open (STDOUT, ">&output_write")  ||
	    die "$PROGRAM [$$]: $!: Redirect stdout to output write " .
		"pipe in child to run: @$argv\n";

	open (STDERR, ">$tmp_error_file")  ||
	    die "$PROGRAM [$$]: $!: Redirect stderr to tmp error " .
		"file: $tmp_error_file\n";

	exec (@$argv)  ||
	    die "$PROGRAM [$$]: $!: Execute command in child: @$argv\n";
    }

    close (output_write)  ||
	die "$PROGRAM [$$]: $!: Close output in father of command " .
	    "run: @$argv\n";

    if ($debug)
    {
	print "[$$]  >>\n",
	      "[$$]  >> $STDOUT_START\n";
    }

    print $client_socket "\n$STDOUT_START\n";

    while (<output_read>)
    {
	chomp;

	if ($debug)
	{
	    print "[$$]  >> $_\n";
	}

	print $client_socket "$_\n";
    }

    waitpid ($pid, 0);

    my $exit_val = $?;

    if ($exit_val < 0)
    {
	$exit_val = 0;
    }

    if ($debug)
    {
	print "[$$]  >> $STDOUT_END\n";
    }

    print $client_socket "$STDOUT_END\n";

    close (output_read)  ||
	die "$PROGRAM [$$]: $!: Close output read pipe in father of " .
	    "command: @$argv\n";

    my $tmp_error_file = getTmpErrorFile ($pid);

    if (-s $tmp_error_file)
    {
	$self->printErrorFileContent ($tmp_error_file, $argv);
    }

    if ($debug)
    {
	print "[$$]  >> $EXIT_VAL_START\n",
	      "[$$]  >> $exit_val\n",
	      "[$$]  >> $EXIT_VAL_END\n";
    }

    print $client_socket
	  "$EXIT_VAL_START\n",
	  "$exit_val\n",
	  "$EXIT_VAL_END\n";

    unlink ($tmp_error_file);
}

sub runWriteWindows
{
    my $self = shift;
    my $argv = shift;

    my $param	      = $self->{$PARAM_KEY	  };
    my $client_socket = $self->{$CLIENT_SOCKET_KEY};

    my $debug = $param->getDebug ();

    my $tmp_error_file = getTmpErrorFile ($$);

    my $win_argv = getWinArgv ($argv, $tmp_error_file);

    local *run_read;

    open (run_read, "@$win_argv|")  ||
	die "$PROGRAM [$$]: $!: Open read pipe to command: " .
	    "@$win_argv\n";

    if ($debug)
    {
	print "[$$]  >>\n",
	      "[$$]  >> $STDOUT_START\n";
    }

    print $client_socket "\n$STDOUT_START\n";

    while (<run_read>)
    {
	chomp;

	if ($debug)
	{
	    print "[$$]  >> $_\n";
	}

	print $client_socket "$_\n";
    }

    my $close_state = close (run_read);
    my $exit_val    = $?;

    if ($exit_val < 0)
    {
	$exit_val = 0;
    }

    if (!$close_state)
    {
	print STDERR
	      "[$$]  $!: Close read pipe to command: @$win_argv\n";
    }

    if ($debug)
    {
	print "[$$]  >> $STDOUT_END\n";
    }

    print $client_socket "$STDOUT_END\n";

    if (-s $tmp_error_file)
    {
	$self->printErrorFileContent ($tmp_error_file, $argv);
    }

    if ($debug)
    {
	print "[$$]  >> $EXIT_VAL_START\n",
	      "[$$]  >> $exit_val\n",
	      "[$$]  >> $EXIT_VAL_END\n";
    }

    print $client_socket
	  "$EXIT_VAL_START\n",
	  "$exit_val\n",
	  "$EXIT_VAL_END\n";

    unlink ($tmp_error_file);
}

sub getWinArgv
{
    my $argv	       = shift;
    my $tmp_error_file = shift;

    my $win_argv = [ 'perl', '-w', '-S', $argv->[$[] . '.pl',
		     '-stderr_file', $tmp_error_file ];

    for (my $index = $[ + 1;  $index <= $#$argv;  $index++)
    {
	push (@$win_argv, $argv->[$index]);
    }

    return $win_argv;
}

sub printErrorFileContent
{
    my $self	       = shift;
    my $tmp_error_file = shift;
    my $argv	       = shift;

    my $client_socket = $self->{$CLIENT_SOCKET_KEY};

    print "[$$]  Command contains \"stderr\" content: @$argv:\n";

    local *tmp_error_read;

    open (tmp_error_read, "<$tmp_error_file")  ||
	die "$PROGRAM [$$]: $!: Open tmp error file for reading: " .
	    "$tmp_error_file\n";

    print $client_socket "$STDERR_START\n";

    while (<tmp_error_read>)
    {
	chomp;

	print $client_socket "$_\n";
	print "            $_\n";
    }

    print $client_socket "$STDERR_END\n";

    close (tmp_error_read)  ||
	die "$PROGRAM [$$]: $!: Close tmp error file after reading: " .
	    " $tmp_error_file\n";
}

sub unlinkTmpErrorFileAndExit
{
    unlink (getTmpErrorFile ($$));
    exit   (1);
}

sub getTmpErrorFile
{
    my $pid = shift;

    return "$TMP_ERROR_BASE.$pid.err";
}

sub closeClient
{
    my $self = shift;

    my $client_socket = $self->{$CLIENT_SOCKET_KEY};
    my $peer_hostname = $self->{$PEER_HOSTNAME_KEY};

    if (!close ($client_socket))
    {
	if ($SYSTEM_TYPE == $UNIX)
	{
	    die "$PROGRAM [$$]: $!: Closing client socket\n";
	}
	else    # $WINDOWS
	{
	    print STDERR "[$$]  $!: Closing client socket\n";
	}
    }

    print "[$$]  Connection $peer_hostname closed\n";
}

sub getTmpErrorBase
{
    if ($SYSTEM_TYPE == $UNIX)
    {
	return "/tmp/$PROGRAM";
    }
    else    # $WINDOWS
    {
	my $tmp_dir = "$PROG_DIR\\..\\Tmp";

	if (! -d $tmp_dir)
	{
	    mkdir ($tmp_dir)  ||
		die "$PROGRAM [$$]: $!: Create tmp directory: " .
		    "$tmp_dir\n";
	}

	return "$tmp_dir\\$PROGRAM";
    }
}

sub dummyFunction
{
    $SIG{CHLD} = \&dummyFunction;    # Required for SysV.
}

package main;

main (\@ARGV);
